(** This module gives an abstract syntax (inspired from logical terms) for affine terms.

    Our terms support a "old" operator that is used to speak about an implicit "old" state. 
    In the usual semantics, the "old" state is identified with the "current" state.

*)

Require Export ProgVar.
Require Export NumC.
Require Import Bool.
Require Import ZArith.
Require Import LinTerm.
Require Export DomainInterfaces.

Create HintDb vpl discriminated.

(* affine terms as expressions *)
Module Term (N: NumSig) <: TermSig N.

  Inductive term: Set :=
  | Var (x: PVar.t)
  | Cte (c: N.t)
  | Plus (tl tr: term)
  | Opp (te: term)
  | Mult (tl:term) (c:N.t)
  | Old (te:term).

  Definition t:=term.
  
  Local Coercion Var: PVar.t >-> term.
  Local Coercion Cte: N.t >-> term.


  (* Standard evaluation (one state) *)
  Fixpoint eval (te:term) (m: Mem.t N.t) : N.t := 
    match te with
      | Var x => m x
      | Cte c => c
      | Plus tl tr => N.add (eval tl m) (eval tr m)
      | Opp te => N.neg (eval te m)
      | Mult tl c => N.mul (eval tl m) c
      | Old te => (eval te m)
    end.

  (* Useless: derived from eval_pointwise_compat ?
  Add Parametric Morphism: eval with
   signature Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> Logic.eq
   as eval_compat.
  Proof.
     intros t m1 m2 H; induction t; simpl; try congruence.
  Qed.
  *)

  (* Extended evaluation (two states) *)
  Fixpoint xeval (te:term) (old new:Mem.t N.t) : N.t := 
    match te with
      | Var x => new x
      | Cte c => c
      | Plus tl tr => N.add (xeval tl old new) (xeval tr old new)
      | Opp te => N.neg (xeval te old new)
      | Mult tl c => N.mul (xeval tl old new) c
      | Old te => eval te old
    end. 

  Lemma xeval_eval te m: xeval te m m = eval te m.
  Proof.
    induction te; simpl; congruence.
  Qed.
  Hint Rewrite xeval_eval: vpl.

  (* Useless: derived from eval_pointwise_compat ?
  Add Parametric Morphism: xeval with
    signature Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> Logic.eq
    as xeval_compat.
  Proof.
     intros t old1 old2 H1 new1 new2 H2; induction t; simpl; try congruence.
     erewrite eval_compat; eauto.
  Qed.
  *)

  (* MayDependOn + mdBound *)
  Fixpoint mayDependOn (te: t) (x: PVar.t): Prop := 
    match te with
      | Var y => x=y 
      | Cte c => False
      | Plus tl tr => mayDependOn tl x \/ mayDependOn tr x
      | Opp te => mayDependOn te x
      | Mult te c => mayDependOn te x 
      | Old te => mayDependOn te x 
    end.

  Hint Resolve (@mdBounds_fold _ mayDependOn): progvar.

  Fixpoint mdBound (te:t) (bound: PVar.t): PVar.t := 
    match te with
      | Var x => PVar.max bound x
      | Cte _ => bound
      | Plus tl tr => mdBound tl (mdBound tr bound)
      | Opp te => mdBound te bound
      | Mult tl c => mdBound tl bound
      | Old te => mdBound te bound
    end.

  Program Instance mdBound_mdBoundVar: MDBoundVar (mayDependOn:=mayDependOn) (mdBound:=mdBound).
  Obligation 1.
    generalize bound; induction e; simpl; intros; autorewrite with rwmax; auto.
    rewrite IHe2; rewrite IHe1. eapply eq_sym; rewrite IHe1.
    autorewrite with rwmax. auto.
  Qed.
  Hint Resolve mdBound_mdBoundVar_obligation_1: progvar.
  Obligation 2.
    unfold mdBounds; generalize bound; induction e; simpl; intros; subst; intuition; eauto with progvar.
  Qed.
  Hint Resolve mdBound_mdBoundVar_obligation_2: progvar.

  Lemma eval_mdo: mdoExt mayDependOn eval Logic.eq.
  Proof.
    unfold mdoExt, bExt; induction e; simpl; intuition auto.
  Qed.

  Lemma xeval_old_mdo old: mdoExt mayDependOn (fun t => xeval t old) Logic.eq.
  Proof.
    unfold mdoExt, bExt; induction e; simpl; intuition.
  Qed.

  Lemma xeval_new_mdo new: mdoExt mayDependOn (fun t old => xeval t old new) Logic.eq.
  Proof.
    unfold mdoExt, bExt; induction e; simpl; intuition.
    erewrite eval_mdo; eauto.
  Qed.

  (* Mapping of variables *)
  Fixpoint map (te: term) (f: Mem.t PVar.t) : term := 
    match te with
      | Var x => Var (f x)
      | Plus tl tr => Plus (map tl f) (map tr f)
      | Opp te => Opp (map te f)
      | Mult te c => Mult (map te f) c
      | Old te => map te f
      | te => te
    end.

  Lemma eval_map: forall te f m, eval (map te f) m = eval te (fun x => m (f x)).
  Proof.
    induction te; simpl; auto.
  Qed.
  Hint Resolve eval_map: vpl.

  Fixpoint xmap (te: term) (old new: Mem.t PVar.t): term := 
    match te with
      | Var x => Var (new x)
      | Plus tl tr => Plus (xmap tl old new) (xmap tr old new)
      | Opp te => Opp (xmap te old new)
      | Mult te c => Mult (xmap te old new) c
      | Old te => Old (map te old)
      | te => te
    end.

  Lemma xeval_xmap: forall te old new f g, xeval (xmap te f g) old new  = xeval te (fun x => old (f x)) (fun x => new (g x)).
  Proof.
    induction te; simpl; auto with vpl.
  Qed.
  Hint Resolve xeval_xmap: vpl.

End Term.

Module Type AffTermSig(N: NumSig).
  Include Term(N).
End AffTermSig.

Module ZTerm := Term ZNum.

Module QTerm.

  Include Term QNum.

  (* translation from ZTerms to QTerms *)
  Fixpoint lift (te: ZTerm.t): term :=  
    match te with
      | ZTerm.Var x => Var x 
      | ZTerm.Cte c => Cte (ZtoQ.ofZ c)
      | ZTerm.Plus tl tr => Plus (lift tl) (lift tr)
      | ZTerm.Opp te => Opp (lift te)
      | ZTerm.Mult tl c => Mult (lift tl) (ZtoQ.ofZ c)
      | ZTerm.Old te => Old (lift te)
    end.

  Lemma lift_correct (te: ZTerm.t) (m: Mem.t ZNum.t):
    eval (lift te) (Mem.lift ZtoQ.ofZ m) = ZtoQ.ofZ (ZTerm.eval te m).  
  Proof.
    induction te; simpl; intros; autorewrite with num; auto.
  Qed.
  Hint Rewrite lift_correct: vpl.


(* Translation from QTerms to LinTerms + coeff *)

  Local Open Scope list_scope.
  Import List.ListNotations.

  Fixpoint toLinRec (te:QTerm.t) (c:QNum.t): Lin.t * QNum.t   := 
    match te with
      | Var x =>  (Lin.single x c,QNum.z)
      | Cte c0 => (Lin.nil,QNum.mul c0 c)
      | Plus tl tr => 
        let (lt1,aux1):=toLinRec tl c in
        let (lt2,aux2):=toLinRec tr c in
        (Lin.add lt1 lt2,QNum.add aux1 aux2)
      | Opp t0 => toLinRec t0 (QNum.neg c)
      | Mult tl c0 => toLinRec tl (QNum.mul c0 c)
      | Old t0 => toLinRec t0 c
    end.

  Add Ring NumRing: QNum.Ring.

  Lemma toLinRec_correct te m: forall c,
      QNum.add (Lin.eval (fst (toLinRec te c)) m) (snd (toLinRec te c)) = (QNum.mul (QTerm.eval te m) c).
  Proof.
    simpl; induction te; simpl; intros; autorewrite with linterm; try (rewrite IHte; simpl); try ring.
    cutrewrite 
      (QNum.mul (QNum.add (eval te1 m) (eval te2 m)) c
       = QNum.add (QNum.mul (eval te1 m) c)
                  (QNum.mul (eval te2 m) c)).
    rewrite <- IHte1, <- IHte2. 
    destruct (toLinRec te1 c), (toLinRec te2 c); simpl.
    autorewrite with linterm.
    ring.
    ring.
  Qed.

  Definition toLin (te: QTerm.t) : Lin.t * QNum.t
    := toLinRec te QNum.u.
  Extraction Inline toLin.

  Lemma toLin_correct te m: 
    QTerm.eval te m = QNum.add (Lin.eval (fst (toLin te)) m) (snd (toLin te)).
  Proof.
    unfold toLin; simpl.
    rewrite toLinRec_correct.
    ring.
  Qed.

End QTerm.
