(* A functor to extend a domain with a "boundVar" operation.

   This operation is computed in "constant" time (if we admit that integers are bounded).

   Except for assume, require, assign, guassign where the term/condition needs to be analyzed,
   all other operations have no additional cost.

*)

Require Export AffCond.
Require Import Debugging.
Require Import PArith.BinPosDef.

Module MakeBoundVarD (N: NumSig) (Import Cond: AffCondSig N) (D: FullAbstractDomain N Cond) <: FullBoundVarDomain N Cond.

  Definition decodeIn (b:PVar.t) (aux m: Mem.t N.t) (x:PVar.t): N.t := 
    if (PVar.isLt b x) then aux x else m x.

  Record boundedD := 
   { raw: D.t ;
     boundVar:  positive 
    }.

  Definition t:=boundedD.

  Definition gamma (a:t) m := forall aux, D.gamma (raw a) (decodeIn (boundVar a) aux m).

  Instance decodeIn_compat : Proper (Logic.eq ==>
    pointwise_relation PVar.t Logic.eq ==>
    pointwise_relation PVar.t Logic.eq ==>
    pointwise_relation PVar.t Logic.eq) decodeIn.
  Proof.
    intros r1 r2 H1 aux1 aux2 H2 m1 m2 H3 x; subst.
    unfold decodeIn. rewrite H2; rewrite H3; auto.
  Qed.

  Instance gamma_ext: Proper (Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> iff) gamma.
  Proof.
    eapply sat_compat_iff; unfold gamma.
    intros x y H m1 m2 H0 H1 aux. rewrite <- H. rewrite <- H0; auto.
  Qed.   


  Definition top: t
    := {| raw := D.top ; boundVar := xH |}.

  Lemma top_correct: forall m, gamma top m.
  Proof.
    unfold gamma; auto with vpl.
  Qed.

  Definition bottom: t
    := {| raw := D.bottom ; boundVar := xH |}.

  Lemma bottom_correct: forall m, gamma bottom m -> False.
  Proof.
    unfold gamma; intros m H; generalize (H m); clear H; eauto with vpl.
  Qed.

  Definition mayDependOn a x := PVar.Le x (boundVar a).

  Lemma gamma_mdo: mdoExt mayDependOn gamma iff.
  Proof.
    assert (mdoExt mayDependOn gamma Program.Basics.impl); auto with progvar.
    unfold gamma, Basics.impl. 
    unfold mdoExt, bExt, mayDependOn, decodeIn. intros e m1 m2 H H0 aux.
    eapply D.gamma_ext; eauto with vpl.
    intro x0. generalize (PVar.LtIsLt (boundVar e) x0).
    case (PVar.isLt (boundVar e) x0); intuition eauto.
    rewrite H; auto. assert (~(PVar.Lt (boundVar e) x0)). 
    intuition.
    PVar.pvar_solve.
  Qed.

  Definition coerce (a:t) (b:PVar.t) : t :=
    {| raw := raw a; boundVar := b |}.

  Lemma coerce_correct a b m: PVar.Le (boundVar a) b -> gamma a m -> gamma (coerce a b) m.
  Proof.
    unfold gamma, coerce. intros H H0 aux; simpl.
    eapply D.gamma_ext; [ eauto | idtac | eapply (H0 (decodeIn b aux m))].
    unfold bExt, decodeIn.
    intros x; generalize (PVar.LtIsLt b x).
    destruct (PVar.isLt b x); intuition.
    + assert (H3:PVar.Lt (boundVar a) x); eauto with progvar.
    rewrite <- PVar.LtIsLt in H3.
    rewrite H3; simpl; auto.
    + destruct (PVar.isLt (boundVar a) x); auto.
  Qed.
  Hint Resolve coerce_correct: progvar.

  Definition join (a1 a2:t) : t :=
    {| raw := D.join (raw a1) (raw a2) ;
      boundVar := PVar.max (boundVar a1) (boundVar a2) |}.

  Lemma join_correct1 a1 a2 m: gamma a1 m -> gamma (join a1 a2) m.
  Proof.
    intros H aux. apply D.join_correct1.
    generalize aux; change (gamma (coerce a1 (PVar.max (boundVar a1) (boundVar a2))) m); eauto with progvar.
  Qed.

  Lemma join_correct2 a1 a2 m: gamma a2 m -> gamma (join a1 a2) m.
  Proof.
    intros H aux. apply D.join_correct2.
    generalize aux; change (gamma (coerce a2 (PVar.max (boundVar a1) (boundVar a2))) m); eauto with progvar.
  Qed.

  Definition widen (a1 a2:t) : t :=
    {| raw := D.widen (raw a1) (raw a2) ;
      boundVar := PVar.max (boundVar a1) (boundVar a2) |}.


(* We choose to restrict isIncl such that [PVar.Le (boundVar a1) (boundVar a2)]
Alternatively, we could compute the bound "p" on a2, and also accept the case
               [PVar.Le p (boundVar a2)]. 
But this seems not necessary because of the [isIncl_relatively_complete] lemma.

Hence, invariant can be systematically tested using [coerce].
*)
  Definition isIncl (a1 a2:t): bool
    := if PVar.isLe (boundVar a1) (boundVar a2)
      then D.isIncl (raw a1) (raw a2)
      else trace INFO "BoundVarD.isIncl: bound of arg2 is too small" false.

  Lemma isIncl_relatively_complete a1 a2: isIncl a1 (coerce a2 (boundVar a1))=D.isIncl (raw a1) (raw a2).
  Proof.
    unfold isIncl, trace; simpl. 
    generalize (PVar.Le_refl (boundVar a1)).
    rewrite <- (PVar.LeIsLe (boundVar a1) (boundVar a1)).
    intros H. rewrite H. auto.
  Qed.

  Lemma isIncl_correct a1 a2: If isIncl a1 a2 THEN forall m, gamma a1 m -> gamma a2 m.
  Proof.
    unfold isIncl; generalize (PVar.LeIsLe (boundVar a1) (boundVar a2)).
    destruct (PVar.isLe (boundVar a1) (boundVar a2)); simpl; auto.
    intros H; VPLsimplify.
    intuition.
    assert (X: gamma (coerce a1 (boundVar a2)) m).
    eauto with progvar.
    unfold gamma in * |- *; simpl in * |- *; eauto.
  Qed.
  Global Opaque isIncl.

  (* NB: the usual implementation of isBottom is not complete ! *)
  Definition isBottom (a:t) : bool :=
    D.isBottom (raw a).

  Lemma isBottom_correct a: If isBottom a THEN forall m, (gamma a m)->False.
  Proof.
    OptionMonad.xstep ltac:(eauto with vpl); auto.
    unfold gamma; intros H m H0.
    generalize (H0 m); clear H0.
    eauto.
  Qed.

  (* Remark: we could avoid to project when "x > boundVar a". But this case may not happen in practice. *)
  Definition project (a:t) (x:PVar.t) : t :=
    {| raw := D.project (raw a) x ; 
      boundVar := if PVar.eq_dec x (boundVar a) then 
        trace INFO "bound--" (Pos.pred x) 
        else (boundVar a) |}.

  Lemma project_correct: forall a x m v, gamma a m -> gamma (project a x) (Mem.assign x v m).
  Proof.
    unfold gamma, project, trace; intros. simpl.
    eapply D.gamma_ext; 
      [ eauto | idtac | eapply D.project_correct ; eauto].
    intros y. unfold decodeIn.
    unfold Mem.assign. case (PVar.eq_dec x y).
    + intros; subst; eauto.
    + intros; case (PVar.eq_dec x (boundVar a)); eauto.
      intros; subst.
      generalize (PVar.LtIsLt (Pos.pred (boundVar a)) y).
      case (PVar.isLt (Pos.pred (boundVar a)) y); intuition.
      - assert (H2: PVar.Lt (boundVar a) y).
        rewrite <- PVar.Le_Lt_pred; eauto.
        rewrite <- PVar.LtIsLt in H2.
        rewrite H2; simpl; eauto.
      - assert (H3: PVar.isLt (boundVar a) y = false).
        assert (H4: ~(PVar.Lt (boundVar a) y)).
        rewrite <- PVar.Le_Lt_pred; intuition.
        rewrite <- PVar.LtIsLt in H4.
        destruct (PVar.isLt (boundVar a) y); intuition.
        rewrite H3; simpl; auto.
  Qed.

  Definition assume (c:Cond.t) (a: t) : t := 
    {| raw := D.assume c (raw a) ; 
      boundVar := mdBound c (boundVar a) |}.

  Lemma assume_correct c a m: (sat c m) -> (gamma a m) -> (gamma (assume c a) m).
  Proof.
    unfold assume; intros H H0 aux; simpl.
    apply D.assume_correct; auto.
    + erewrite sat_mdo; eauto.
    intros x H1; unfold decodeIn.
    generalize (PVar.LtIsLt (mdBound c (boundVar a)) x).
    case (PVar.isLt (mdBound c (boundVar a)) x); simpl; intuition.
    case (md_mdBounds _ _ H1 (mdBound c (boundVar a))); eauto with progvar.
    + generalize aux; change (gamma (coerce a (mdBound c (boundVar a))) m).
    eauto with progvar.
  Qed.

  Definition require (c:Cond.t) (a: t): bool := 
    D.require c (raw a).

  Lemma require_correct c a: If require c a THEN forall m, gamma a m -> sat c m. 
  Proof.
    VPLsimplify.
    intros H m H0; assert (X: gamma (coerce a (mdBound c (boundVar a))) m); eauto with progvar.
    clear H0; erewrite sat_mdo; eauto.
    intros x H1; unfold decodeIn; simpl.
    generalize (PVar.LtIsLt (mdBound c (boundVar a)) x).
    case (PVar.isLt (mdBound c (boundVar a)) x); simpl; intuition.
  Qed.

  Definition assign x (te:Term.t) (a: t) : t := 
    {| raw := D.assign x te (raw a) ; 
      boundVar := Term.mdBound te (PVar.max x (boundVar a)) 
      |}.

  Lemma assign_correct x te a m: 
    gamma a m -> gamma (assign x te a) (Mem.assign x (Term.eval te m) m).
  Proof.
    intros H; assert (H0: gamma (coerce a (Term.mdBound te (PVar.max x (boundVar a)))) m).
    eauto with progvar.
    clear H; unfold gamma, assign in * |- *; simpl in * |- *; intro aux.
    eapply D.gamma_ext; 
      [ eauto | idtac | eapply D.assign_correct; eauto ].
    intros y; unfold Mem.assign, decodeIn; simpl.
    case (PVar.eq_dec x y); auto.
    intros; subst. assert (PVar.Le y (Term.mdBound te (PVar.max y (boundVar a)))).
    eauto with progvar.
    assert (X:~(PVar.Lt (Term.mdBound te (PVar.max y (boundVar a)))) y).
    PVar.pvar_solve.
    rewrite <- PVar.LtIsLt in X.
    destruct (PVar.isLt (Term.mdBound te (PVar.max y (boundVar a))) y); intuition.
    erewrite Term.eval_mdo; eauto.
    intros x H1. generalize (PVar.LtIsLt (Term.mdBound te (PVar.max y (boundVar a))) x).
    case (PVar.isLt (Term.mdBound te (PVar.max y (boundVar a))) x); auto.
    intro H2; case (md_mdBounds _ _ H1 (Term.mdBound te (PVar.max y (boundVar a)))); intuition.
  Qed.    

  (* guarded assignement *)
  Definition guassign x c a : t :=
    {| raw := D.guassign x c (raw a) ; 
      boundVar := mdBound c (PVar.max x (boundVar a)) 
      |}.

  Lemma guassign_correct x c a m v:
    (xsat c m (Mem.assign x v m)) -> gamma a m -> gamma (guassign x c a) (Mem.assign x v m).
  Proof.
    intros H H0; assert (H1: gamma (coerce a (mdBound c (PVar.max x (boundVar a)))) m).
    eauto with progvar.
    clear H0; unfold gamma, assign in * |- *; simpl in * |- *; intro aux.
    eapply D.gamma_ext; 
      [ eauto | idtac | eapply D.guassign_correct; [ idtac | eauto ]].
    + intros y; unfold Mem.assign, decodeIn; simpl.
      case (PVar.eq_dec x y); auto.
      intros; subst. assert (PVar.Le y (mdBound c (PVar.max y (boundVar a)))).
      eauto with progvar.
      assert (X:~(PVar.Lt (mdBound c (PVar.max y (boundVar a)))) y).
      PVar.pvar_solve.
      rewrite <- PVar.LtIsLt in X.
      destruct (PVar.isLt (mdBound c (PVar.max y (boundVar a))) y); [ case X; auto | eauto ].
    + clear H1.
      erewrite Cond.xsat_old_mdo.
      erewrite Cond.xsat_new_mdo; eauto.
      * unfold decodeIn; intros y H1. 
        generalize (PVar.LtIsLt (mdBound c (PVar.max x (boundVar a))) y).
        case (PVar.isLt (mdBound c (PVar.max x (boundVar a))) y); auto.
        intro H2; case (md_mdBounds _ _ H1 (mdBound c (PVar.max x (boundVar a)))); intuition.
      * unfold Mem.assign, decodeIn;  simpl.
        intros y H1; case (PVar.eq_dec x y); auto.
        intros; generalize (PVar.LtIsLt (mdBound c (PVar.max x (boundVar a))) y).
        case (PVar.isLt (mdBound c (PVar.max x (boundVar a))) y); auto.
        intros; case (md_mdBounds c y H1 (mdBound c (PVar.max x (boundVar a)))); intuition.
  Qed.

End MakeBoundVarD. 
