(** This module implements the [AbIdealEnv.ab_ideal_env] interface
of abtract domains on ideal numbers. *)

Require Import ZArith_base.
Require PolZ.
Import NumC ProgVar AffExpr Itv.
Require Import LatticeSignatures.
Require Import AdomLib.
Require Import AbMachineEnv.
Require Import AbIdealEnv.
Require Import IdealEnvToMachineEnv.
Require Import String.
Require Import ToString.

Local Open Scope Z_scope.

Local Notation var := positive.

Definition liftN (dummy: ZNum.t) (n: ideal_num): ZNum.t
  := match n with
         | INf _ => dummy
         | INz z => z
     end.

Definition liftRho (rho: var -> ideal_num) (dummy: ZNum.t): ZMem.t
  := fun x => liftN dummy (rho x).

Definition CmpTransSpec (cmp: Integers.comparison) (cmp': ZExpr.cmpT): Prop
  := forall z1 z2: Z, ZIntervals.Zcmp cmp z1 z2 = ZExpr.cmpEval cmp' z1 z2.

(* XXX: should support for "not equal" be added to PolZ? *)
Definition cmpTrans (cmp: Integers.comparison):
  Exc {cmp': ZExpr.cmpT | CmpTransSpec cmp cmp'}.
  refine
    (match cmp with
       | Integers.Ceq => value (exist _ ZExpr.Eq _)
       | Integers.Cne => error
       | Integers.Clt => value (exist _ ZExpr.Lt _)
       | Integers.Cle => value (exist _ ZExpr.Le _)
       | Integers.Cgt => value (exist _ ZExpr.Gt _)
       | Integers.Cge => value (exist _ ZExpr.Ge _)
     end);
  unfold CmpTransSpec;
  unfold ZIntervals.Zcmp;
  intros z1 z2;
  match goal with
    | |- match Z.compare z1 z2 with Eq => ?a | Lt => ?b | Gt => ?c end = ?d =>
      replace (match Z.compare z1 z2 with
                 | Eq => a
                 | Lt => b
                 | Gt => c end = d)
      with (match Z.compare z1 z2 with
              | Eq => a = d
              | Lt => b = d
              | Gt => c = d end);
        [ |destruct (Z.compare z1 z2);
            reflexivity]
    | _ => idtac
  end;
  apply Zcompare_elim;
    intro h;
    try reflexivity;
  symmetry;
  match goal with
    | |- _ = true => apply ZExpr.CmpEvalDenoteT
    | |- _ = false => apply ZExpr.CmpEvalDenoteF
  end;
  simpl.
  - assumption.
  - intro hcontr; rewrite hcontr in h.
    apply Z.lt_irrefl in h.
    destruct h.
  - intro hcontr; rewrite hcontr in h.
    apply Zgt_irrefl in h.
    destruct h.
  - intro hcontr; rewrite h in hcontr.
    apply Zlt_irrefl in hcontr.
    destruct hcontr.
  - assumption.
  - apply Zgt_not_le in h.
    intro hcontr.
    apply Zlt_le_weak in hcontr.
    destruct h.
    assumption.
  - rewrite h.
    apply Zle_refl.
  - apply Zlt_le_weak in h.
    assumption.
  - intro hcontr.
    apply Zle_not_gt in hcontr.
    destruct hcontr.
    assumption.
  - rewrite h.
    intro hcontr.
    apply Zlt_irrefl in hcontr.
    destruct hcontr.
  - apply Zlt_not_le in h.
    intro hcontr.
    apply Zlt_le_weak in hcontr.
    destruct h.
    assumption.
  - apply Zgt_lt in h.
    assumption.
  - rewrite h.
    apply Zle_refl.
  - apply Zlt_gt in h.
    apply Zgt_not_le in h.
    assumption.
  - apply Zgt_lt in h.
    apply Zlt_le_weak in h.
    assumption.
Defined.

Definition ItvTrSpec (i1: ZIntervals.itv) (i2: ZItv.t): Prop
  := forall z: Z.t, ZIntervals.gamma i1 z <-> ZItv.sat i2 z.

Definition itvTr (i1: ZIntervals.itv): {i2: ZItv.t | ItvTrSpec i1 i2}.
  refine (
      match i1
            return {i2: ZItv.t | ItvTrSpec i1 i2}
      with
        | ZIntervals.Top => exist _ (proj1_sig (ZItv.mk ZItv.Infty ZItv.Infty)) _
        | ZIntervals.ITV z1 z2 =>
          exist _ (proj1_sig (ZItv.mk (ZItv.Closed z1) (ZItv.Closed z2))) _
      end);
  unfold ItvTrSpec;
  intro z;
  split;
  intro h;
  simpl.
  - repeat constructor.
  - constructor.
  - unfold ZIntervals.gamma in h.
    destruct (ZNum.ltLeDec z2 z1) as [h1|h1].
    + assert (h2 := h1). (* duplicate because h1 is used in goal *)
      apply Zlt_not_le in h2.
      destruct h2.
      apply Z.le_trans with (m := z);
        apply h.
    + simpl.
      assumption.
  - simpl in h.
    destruct (ZNum.ltLeDec z2 z1) as [h1|h1].
    + destruct h.
    + simpl in h.
      assumption.
Defined.

Definition IExprTrSpec (ie: iexpr var) (e: PolZ.exprT): Prop
  := forall (dummy: ZNum.t) (n: ideal_num) (m: var -> ideal_num),
       eval_iexpr m ie n -> PolZ.EvalExpr (liftRho m dummy) e (liftN dummy n).

Fixpoint iexprTr (ie: iexpr var): {e: PolZ.exprT | IExprTrSpec ie e}.
  refine (
      match ie return {e: PolZ.exprT | IExprTrSpec ie e} with
        | IEvar x => exist _ (PolZ.Evar x) _
        | IEZconst i => exist _ (PolZ.Econst (proj1_sig (itvTr i))) _
        | IEFconst _ => exist _ PolZ.Eunknown _
        | IEUnknown => exist _ PolZ.Eunknown _
        | IEunop op ie' =>
          let (e', pfe) := iexprTr ie' in
          match op
                return {e: PolZ.exprT | IExprTrSpec (IEunop op ie') e}
          with
            | IOneg => exist _ (PolZ.Eunop PolZ.Oneg e') _
            | IOnot => exist _ (PolZ.Eunop PolZ.Onot e') _
            | _ => exist _ PolZ.Eunknown _
          end
        | IEbinop op ie1 ie2 =>
          let (e1, pfe1) := iexprTr ie1 in
          let (e2, pfe2) := iexprTr ie2 in
          match op
                return {e: PolZ.exprT | IExprTrSpec (IEbinop op ie1 ie2) e}
          with
            | IOadd => exist _ (PolZ.Ebinop PolZ.Oadd e1 e2) _
            | IOsub => exist _ (PolZ.Ebinop PolZ.Osub e1 e2) _
            | IOmul => exist _ (PolZ.Ebinop PolZ.Omul e1 e2) _
            | IOdiv => exist _ (PolZ.Ebinop PolZ.Odiv e1 e2) _
            | IOmod => exist _ (PolZ.Ebinop PolZ.Omod e1 e2) _
            | IOand => exist _ (PolZ.Ebinop PolZ.Oand e1 e2) _
            | IOor => exist _ (PolZ.Ebinop PolZ.Oor e1 e2) _
            | IOxor => exist _ (PolZ.Ebinop PolZ.Oxor e1 e2) _
            | IOshl => exist _ (PolZ.Ebinop PolZ.Oshl e1 e2) _
            | IOshr => exist _ (PolZ.Ebinop PolZ.Oshr e1 e2) _
            | IOcmp c =>
              match cmpTrans c with
                | error => exist _ PolZ.Eunknown _
                | value (exist c' pfc) =>
                  exist _ (PolZ.Ebinop (PolZ.Ocmp c') e1 e2) _
              end
            | _ => exist _ PolZ.Eunknown _
          end
      end);
  unfold IExprTrSpec;
  intros dummy n m h;
  inversion h;
  try constructor;
  match goal with
    | hyp1: eval_iexpr _ _ ?n0,
            hyp2: eval_iunop _ ?n0 _
      |- PolZ.EvalExpr _ (PolZ.Eunop _ _) _
      => apply PolZ.EvalEunop with (z1 := liftN dummy n0);
        [apply pfe; assumption|];
        match goal with
          | hyp: eval_iunop _ n0 n |- _ => inversion hyp
        end;
        constructor

    | hyp1: eval_iexpr _ ?ie1 ?n1,
            hyp2: eval_iexpr _ ?ie2 ?n2,
                  hyp3: eval_ibinop ?op ?n1 ?n2 _
      |- PolZ.EvalExpr _ (PolZ.Ebinop _ _ _) _
      => apply PolZ.EvalEbinop with (z1 := liftN dummy n1) (z2 := liftN dummy n2);
        [apply pfe1; assumption|apply pfe2; assumption|idtac];
        match goal with
          | hyp: eval_ibinop _ _ _ _ |- _ => inversion hyp
        end;
        match goal with
          | hyp: CmpTransSpec _ _|- PolZ.EvalBinop (PolZ.Ocmp _) _ _ _
            => rewrite hyp
          | _ => idtac
        end;
        constructor

    | _ => idtac
  end.
  - unfold liftRho.
    match goal with
      | hyp: m x = n |- _ => rewrite hyp
    end.
    reflexivity.
  - simpl.
    apply (proj2_sig (itvTr i)).
    assumption.
  - assumption.
  - assumption.
Defined.

Definition A: Type
  := PolZ.A.

Definition B: Type
  := var -> ideal_num.

Instance PolWeakLattice: weak_lattice A
  := {
      leb := PolZ.leb;
      top := PolZ.top;
      join := PolZ.join;
      widen := PolZ.widen
    }.

Definition gamma: A -> B -> Prop
  := fun ab rho => PolZ.gamma ab (liftRho rho ZNum.z).

Instance PolGammaOp: gamma_op A B
  := {
      γ := gamma
     }.

Lemma gamma_monotone: forall (ab1 ab2: A),
                        leb ab1 ab2 = true ->
                        forall rho: B, gamma ab1 rho -> gamma ab2 rho.
  unfold gamma.
  intros ab1 ab2 h rho.
  apply PolZ.gamma_monotone.
  assumption.
Qed.

Lemma gamma_top: forall x: B, gamma top x.
  unfold gamma.
  intro x.
  generalize (liftRho x ZNum.z).
  apply PolZ.gamma_top.
Qed.

Lemma join_sound: forall (ab1 ab2: A) (x: B),
    gamma ab1 x \/ gamma ab2 x -> gamma (join ab1 ab2) x.
  intros ab1 ab2 x H.
  destruct H;
    apply PolZ.join_sound;
    [left|right];
    assumption.
Qed.

Lemma PolAdom: adom A B PolWeakLattice PolGammaOp.
  split.
  exact gamma_monotone.
  exact gamma_top.
  exact join_sound.
Qed.

Definition ItvConvSpec (i1: ZItv.t) (i2: ZIntervals.itv+⊥): Prop
  := forall z: BinInt.Z, ZItv.sat i1 z -> γ i2 z.

Definition itvConv (i1: ZItv.t): {i2: ZIntervals.itv+⊥ | ItvConvSpec i1 i2}.
  refine (
      match i1
            return {i2: ZIntervals.itv+⊥ | ItvConvSpec i1 i2}
      with
        | ZItv.Bot => exist _ Bot _
        | ZItv.NotBot low up pfi1 =>
          match low
                  return forall pf: ZItv.BndOk low up,
                           {i2: ZIntervals.itv+⊥ | ItvConvSpec (ZItv.NotBot low up pf) i2}
          with
            | ZItv.Infty => fun pfl => exist _ (NotBot ZIntervals.Top) _
            | ZItv.Open b1 =>
              fun pfl =>
                match up
                      return forall pf: ZItv.BndOk (ZItv.Open b1) up,
                               {i2: ZIntervals.itv+⊥ |
                                ItvConvSpec (ZItv.NotBot (ZItv.Open b1) up pf) i2}
                with
                  | ZItv.Infty => fun pfu => exist _ (NotBot ZIntervals.Top) _
                  | ZItv.Open b2
                  | ZItv.Closed b2 =>
                    fun pfu => exist _ (AdomLib.NotBot (ZIntervals.ITV b1 b2)) _
                end pfl
            | ZItv.Closed b1 =>
              fun pfl =>
                match up
                      return forall pf: ZItv.BndOk (ZItv.Closed b1) up,
                               {i2: ZIntervals.itv+⊥ |
                                ItvConvSpec (ZItv.NotBot (ZItv.Closed b1) up pf) i2}
                with
                  | ZItv.Infty => fun pfu => exist _ (NotBot ZIntervals.Top) _
                  | ZItv.Open b2
                  | ZItv.Closed b2 =>
                    fun pfu => exist _ (NotBot (ZIntervals.ITV b1 b2)) _
                end pfl
          end pfi1
      end);
  unfold ItvConvSpec;
  intros z h;
  unfold γ;
  simpl;
  match goal with
    | |- γ ZIntervals.Top z => constructor
    | _ => idtac
  end;
  unfold γ;
  simpl;
  unfold ZItv.sat, ZItv.satLower, ZItv.satUpper in h;
  match goal with
    | h: False |- False => destruct h
    | _ => idtac
  end.
  - split;
      apply Zlt_le_weak, h.
  - split;
    [apply Zlt_le_weak|idtac];
    apply h.
  - split;
    [idtac|apply Zlt_le_weak];
    apply h.
  - split;
    apply h.
Defined.

Definition get_itv: iexpr var -> A -> ZIntervals.itv+⊥
  := fun e ab =>
       let (e', pfe) := iexprTr e in
       proj1_sig (itvConv (PolZ.get_itv e' ab)).

Lemma get_itv_correct: forall (e: iexpr var) (rho: B) (ab: A),
                         gamma ab rho ->
                         (eval_iexpr rho e ∘ INz) ⊆ γ (get_itv e ab).
  simpl.
  intros e rho ab hab z h.
  unfold get_itv.
  destruct (iexprTr e) as (e', pfe).
  apply (proj2_sig (itvConv _)).
  apply PolZ.get_itv_correct with (rho := liftRho rho 0).
  assumption.
  apply pfe with (n := INz z).
  assumption.
Qed.

(* XXX: the bottom handling should be removed from PolC. *)
Definition PropagateBotSpec (ab1 ab2: A+⊥): Prop
  := forall rho: B, lift_gamma gamma ab1 rho <-> lift_gamma gamma ab2 rho.

Definition propagateBot (ab1: A+⊥): {ab2: A+⊥ | PropagateBotSpec ab1 ab2}.
  refine (
      match ab1
            return {ab2: A+⊥ | PropagateBotSpec ab1 ab2}
      with
        | Bot => exist _ Bot _
        | NotBot ab' =>
          match ab'
                return {ab2: A+⊥ | PropagateBotSpec (NotBot ab') ab2}
          with
            | PolC.Bottom => exist _ Bot _
            | PolC.NonBot p => exist _ (NotBot (PolC.NonBot p)) _
          end
      end);
  unfold PropagateBotSpec;
  intros rho;
  split;
  intro h;
  simpl;
  simpl in h;
  assumption || destruct h.
Defined.

Definition assumeAux: iexpr var -> bool -> A -> A+⊥
  := fun e b ab =>
       let (e', pfe) := iexprTr e in
       NotBot (PolZ.assume e' b ab).

Definition assume: iexpr var -> bool -> A -> A+⊥
  := fun e b ab => proj1_sig (propagateBot (assumeAux e b ab)).

Lemma assume_correct: forall (c: iexpr var) (rho: B) (ab: A) (b: bool),
                        gamma ab rho ->
                        eval_iexpr rho c (INz(if b then 1 else 0)) ->
                        lift_gamma gamma (assume c b ab) rho.
  intros e rho ab b hab hc.
  unfold assume.
  apply (proj2_sig (propagateBot (assumeAux e b ab))).
  unfold assumeAux.
  destruct (iexprTr e) as (e', pfe).
  simpl.
  apply PolZ.assume_correct.
  assumption.
  apply pfe with (n := INz (if b then 1 else 0)).
  assumption.
Qed.

Definition assignAux: var -> iexpr var -> A -> A+⊥
  := fun x e ab =>
       let (e', pfe) := iexprTr e in
       NotBot (PolZ.assign x e' ab).

Definition assign: var -> iexpr var -> A -> A+⊥
  := fun x e ab => proj1_sig (propagateBot (assignAux x e ab)).

Instance VarEqDec: EqDec var
  := { eq_dec := PosEqDec }.

Lemma assign_correct: forall (x: var) (e: iexpr var) (ρ: var -> ideal_num)
                             (n: ideal_num) (ab: A),
                        γ ab ρ -> eval_iexpr ρ e n ->
                        γ (assign x e ab) (upd ρ x n).
  intros x e rho n ab h1 h2.
  unfold assign.
  apply (proj2_sig (propagateBot (assignAux x e ab))).
  unfold assignAux.
  destruct (iexprTr e) as (e', pfe).
  simpl.
  assert (pf := PolZ.assign_correct).
  specialize (pf x e' (liftRho rho 0) (liftN 0 n) ab).
  unfold γ, gamma.
  apply PolC.satExt with (m := Mem.liftZMem
                                 (ZMem.assign (liftRho rho 0) x (liftN 0 n))).
  - unfold Mem.Equiv. (* XXX: should be done elsewhere *)
    unfold ZMem.assign.
    unfold liftRho, upd.
    unfold Mem.liftZMem.
    unfold eq_dec.
    intro x0.
    destruct (VarEqDec x0 x) as [pfeq|pfeq].
    + replace (PVar.isEq x x0) with true.
      reflexivity.
      symmetry.
      apply PVar.EqIsEq.
      symmetry.
      assumption.
    + replace (PVar.isEq x x0) with false.
      reflexivity.
      assert (h: ~ PVar.isEq x x0 = true).
      intro hcontr.
      apply PVar.EqIsEq in hcontr.
      destruct pfeq.
      symmetry.
      assumption.
      destruct (PVar.isEq x x0);
        [destruct h|];
        reflexivity.
  - apply pf.
    assumption.
    apply pfe.
    assumption.
Qed.

Definition to_string: A -> string
  := fun p => PolZ.pr p.

Instance PolToString: ToString A
  := {
      to_string := to_string
    }.

Instance PolAbIdealEnv: ab_ideal_env (var := positive) A (H := PosEqDec)
  := {
      id_wl := PolWeakLattice;
      id_gamma := PolGammaOp;
      id_adom := PolAdom;
      get_itv := get_itv;
      assign := assign;
      assume := assume;
      get_itv_correct := get_itv_correct;
      assign_correct := assign_correct;
      assume_correct := assume_correct
    }.

Instance PolAbMachEnv: ab_machine_env A
  := idealenv_to_machineenv PolAbIdealEnv.
