(** This module implements the basic abstract domain of polyhedra
on affine conditions over rationals ([AffCond.QCond]).

It builds on [ConsSet] and adds to it the coupling with opaque
Ocaml polyhedra, operations on which serve as an oracle for
operations in [ConsSet]. It also adds the possibility for a
polyhedron to be empty. 

*)

Require Import List.
Require Import String.
Require CoqAddOn.
Require Import LinTerm.
Require Import ConsSet.
Require Shadow.
Require Import Itv.
Require Export AffCond.
Require Import Debugging.
Require Import AssignD.

Import QCond.
Open Scope option.

Module BasicPolQ <: BasicDomain QNum.

(** Coupling of a [ConsSet] with an opaque Ocaml polyhedron. *)
  Record polT: Type := mk {
    cons: Cs.t;  (* NB. implicit invariant: cons is not empty ! *)
    ml: Shadow.t
  }.

  Definition polPr: polT -> string
    := fun p =>
      CoqAddOn.sprintf "{coq:\n%s\nocaml:\n%s}"
      ((Cs.pr (cons p))::(Shadow.pr (ml p))::nil).

  Definition polTop: polT
    := mk Cs.top Shadow.top.

(** Empty polyhedron is encoded as [None], 
    [Some p] encodes a non-empty polyhedron
*)
  Definition t:=option polT.

  Local Open Scope string_scope.
  Definition pr: t -> string
    := fun p =>
      match p with
        | None => "bot"
        | Some p' => "notbot " ++ (polPr p')
      end.


  Definition gamma (a: t) (m: Mem.t QNum.t) :=
    EXISTS p <- a SUCH Cs.sat (cons p) m.

   Instance gamma_ext: Proper (Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> iff) gamma.
   Proof.
     eapply sat_compat_iff; unfold gamma.
     intros x y H m1 m2 H0 H1. subst; PolCsimplify. 
     erewrite (eval_pointwise_compat Cs.sat_mdo); eauto.
   Qed.

  Definition top: t
    := Some polTop.

  Lemma top_correct: forall m, gamma top m.
    simpl; auto.
  Qed.
  Hint Resolve top_correct: vpl.

  Definition bottom: t := None.

  Lemma bottom_correct: forall m, ~(gamma bottom m).
  Proof.
    simpl; auto.
  Qed.

  Definition isIncl (p1 p2:t) : bool
    :=  match p1 with
          | None => true
          | Some pol1 =>
            match p2 with
              | None => false  (* implicit invariant p1 is not empty *)
              | Some pol2 =>
                match Shadow.isIncl (ml pol1) (ml pol2) with
                  | Shadow.NotIncl => false
                  | Shadow.Incl cert => Cs.chkIncl (cons pol1) (cons pol2) cert
                end
            end
        end.

  Opaque Cs.chkEmpty Cs.chkIncl Cs.redAdd.

  Lemma isIncl_correct p1 p2: If isIncl p1 p2 THEN forall m, gamma p1 m -> gamma p2 m.
  Proof.
    unfold isIncl.
    destruct p1; simpl; intuition.
    destruct p2; simpl; auto.
    destruct (Shadow.isIncl (ml p) (ml p0)); simpl; auto.
    PolCsimplify.
  Qed.
  Hint Resolve isIncl_correct: vpl.
  Opaque isIncl.


  Definition meet (p1 p2:t): t :=
    DO p1' <- p1 -;
    DO p2' <- p2 -;
    let (maybeP, ce) := Shadow.meet (ml p1') (ml p2') in
      match maybeP with
        | Some p => Some (mk (Cs.meet (cons p1') (cons p2') ce) p) 
        | None =>
          if Cs.meetEmpty (cons p1') (cons p2') ce
            then None
            else failwith "PolC.meet" top
      end.

  Lemma meet_correct p1 p2 m: gamma p1 m -> gamma p2 m -> gamma (meet p1 p2) m.
  Proof.
    unfold meet. PolCsimplify.
  Qed.

  Definition join (p1 p2:t) : t := 
    match p1, p2 with
      | None, None => None
      | None, pol
      | pol, None => pol
      | Some pol1, Some pol2 =>
        let (shadow, cert) := Shadow.join (ml pol1) (ml pol2) in
          Some (mk (Cs.buildJoin (cons pol1) (cons pol2) cert) shadow)
    end.

  Lemma join_correct1 p1 p2 m: gamma p1 m -> gamma (join p1 p2) m.
  Proof.
    unfold join; PolCsimplify.
  Qed.

  Lemma join_correct2 p1 p2 m: gamma p2 m -> gamma (join p1 p2) m.
  Proof.
    unfold join; PolCsimplify.
  Qed.
  Hint Resolve join_correct1 join_correct2: vpl.

  Definition widen (p1 p2: t): t :=
    match p1, p2 with
      | None, None => None
      | None, p | p, None => p
      | Some pol1, Some pol2 =>
        let (cs, shadow) := Shadow.widen (ml pol1) (ml pol2) in
          Some (mk shadow cs)
    end.


  Definition project (p: t) (x: PVar.t): t
    := DO p1 <- p -;
       let (p2, c) := Shadow.project (ml p1) x in
       let cs:= Cs.approx (cons p1) c in
        if Cs.isFree x cs then
           Some (mk cs p2)
         else
           failwith ("PolC.project: "++(PVar.pr x)++" not free in "++(Cs.pr cs)) top.

  Lemma project_correct: forall p x m v, gamma p m -> gamma (project p x) (Mem.assign x v m).
  Proof.
    unfold project; PolCsimplify.
    intros; erewrite (mdoExt_free Cs.sat_mdo); eauto with polC progvar.
  Qed.

End BasicPolQ.

Module PolQItv <: HasGetItv QNum QTerm QItv BasicPolQ.

  Import BasicPolQ.

  Definition buildLow (p: polT) (v: Lin.t) (sItv: Shadow.itvT): QItv.bndT :=
      match Shadow.low sItv with
        | Shadow.Infty => QItv.Infty
        | Shadow.Open n f =>
          let c:=Cs.combine (cons p) f in
            if Cstr.isIncl c (Cstr.lowerToCstr v n) then QItv.Open n else QItv.Infty
        | Shadow.Closed n f =>
          let c:=Cs.combine (cons p) f in
            if Cstr.isIncl c (Cstr.lowerOrEqualsToCstr v n) then QItv.Closed n else QItv.Infty
      end.

  Local Opaque Cstr.isIncl.

  Lemma buildLow_correct (p: polT) (v: Lin.t) (sItv: Shadow.itvT) m: 
    gamma (Some p) m -> QItv.satLower (buildLow p v sItv) (Lin.eval v m).
  Proof.
    unfold buildLow; destruct (Shadow.low sItv); simpl; auto; PolCsimplify.
  Qed.

  Definition buildUp (p: polT) (v: Lin.t) (sItv: Shadow.itvT): QItv.bndT :=
      match Shadow.up sItv with
        | Shadow.Infty => QItv.Infty
        | Shadow.Open n f =>
          let c:=Cs.combine (cons p) f in
            if Cstr.isIncl c (Cstr.upperToCstr v n) then QItv.Open n else QItv.Infty
        | Shadow.Closed n f =>
          let c:=Cs.combine (cons p) f in
            if Cstr.isIncl c (Cstr.upperOrEqualsToCstr v n) then QItv.Closed n else QItv.Infty
      end.

  Lemma buildUp_correct (p: polT) (v: Lin.t) (sItv: Shadow.itvT) m: 
    gamma (Some p) m -> QItv.satUpper (buildUp p v sItv) (Lin.eval v m).
  Proof.
    unfold buildUp; destruct (Shadow.up sItv); simpl; auto; PolCsimplify.
  Qed.
  
  (** wrapping "mk" function of QItv *)
  Definition QItv_mk (l u: QItv.bndT): QItv.t
   := projT1 (QItv.mk l u).

  Lemma QItv_mk_correct l u x: QItv.satLower l x -> QItv.satUpper u x -> QItv.sat (QItv_mk l u) x.
  Proof.
    unfold QItv_mk. destruct (QItv.mk l u); simpl; auto.
  Qed.

  Definition getItv (p: t) (v: Lin.t): QItv.t := 
    match p with
    | None => QItv.Bot
    | Some pol =>
      let sItv := Shadow.getItv (ml pol) v in
      let low := buildLow pol v sItv in
      let up := buildUp pol v sItv in
          QItv_mk low up
    end.

  Local Hint Resolve buildUp_correct buildLow_correct  QItv_mk_correct: polC.

  Lemma getItv_correct (p: t) (v: Lin.t) m: gamma p m -> QItv.sat (getItv p v) (Lin.eval v m).
  Proof.
    unfold getItv; PolCsimplify.
  Qed.

  Local Hint Resolve getItv_correct: polC.

  Definition get_itv: QTerm.t -> t -> QItv.t
    := fun a ab =>
      let (t,n) := QTerm.toLin a in
      QItv.shift (getItv ab t) n.

  Lemma get_itv_correct: forall (a: QTerm.t) (rho: Mem.t QNum.t) (ab: t),
    gamma ab rho ->
    QItv.sat (get_itv a ab) (QTerm.eval a rho).
  Proof.
    intros a rho ab hab; unfold get_itv.
    PolCsimplify.
    rewrite QTerm.toLin_correct.
    rewrite QNum.AddComm.
    apply QItv.ShiftsatSpec.
    auto with polC.
  Qed.

End PolQItv.


Module CstrPol <: HasConcreteConditions QNum Cstr BasicPolQ.

  Import BasicPolQ.

  Definition assume (c: Cstr.t) (p: t): t :=
    DO p1 <- trace DEBUG "in CstrPol.assume" p  -;
    let (opt, cert) := Shadow.add (ml p1) c in
      match opt with
        | None =>
          if (Cs.chkEmpty (Cs.add (cons p1) (Shadow.getId (ml p1)) c) cert) then 
            None
          else
            failwith "CstrPol.assume" top
        | Some s => Some (mk (Cs.redAdd (cons p1) (Shadow.getId (ml p1)) c cert) s)
      end.
  
  Lemma assume_correct: forall c a m, Cstr.sat c m -> gamma a m -> gamma (assume c a) m.
  Proof.
    unfold assume. PolCsimplify. simpl in * |- *.
    unfold not; eauto with polC.
  Qed. 

(** [fromCstr c] builds a polyhedron from a constraint.
If it fails, [default] is returned. 
Hence [default] may be chosen according to the fact that precision or soundness is required *)
  Definition fromCstr: Cstr.t -> t -> t
    := fun c default =>
         match Shadow.addp Shadow.top c with
           | Shadow.Added p (fwd, bwd) =>
             match Cs.addp Cs.top (Shadow.getId Shadow.top) c fwd bwd with
               | Some cs => Some (mk cs p)
               | None => failwith "CstrPol.fromCstr" default
             end
           | Shadow.Contrad ce =>
             if Cs.chkEmpty (Cs.add Cs.top (Shadow.getId Shadow.top) c) ce
             then None
             else failwith "CstrPol.fromCstr" default
         end.

(** Currently useless ??  TODO: adapt with [default] 
  Lemma fromCstr_correct: forall c p, fromCstr c = Some p ->
                                      forall m, Cstr.sat c m -> gamma p m.
  Proof.
    intros c p hr m hc.
    unfold fromCstr in hr.
    destruct (Shadow.addp Shadow.top c) as [p' [fwd bwd]|ce].
    - assert (pf := fun cs' =>
                      Cs.addp_correct Cs.top cs' (Shadow.getId Shadow.top)
                                      c fwd bwd m).
      destruct (Cs.addp Cs.top (Shadow.getId Shadow.top) c fwd bwd) as [cs|].
      + inversion hr.
        apply pf; [reflexivity|constructor|assumption].
      + discriminate hr.
    - assert (pf := Cs.chkEmpty_correct
                      (Cs.add Cs.top (Shadow.getId Shadow.top) c) ce).
      destruct (Cs.chkEmpty (Cs.add Cs.top (Shadow.getId Shadow.top) c) ce).
      + inversion hr.
        contradiction pf with (m := m).
        apply Cs.AddsatSpec; [constructor|assumption].
      + discriminate hr.
  Qed.
*)
  Local Opaque Cs.chkEmpty Cs.addp.
  Local Hint Resolve Cs.addp_precise_c: polC.

  Lemma fromCstr_precise: forall c m, gamma (fromCstr c None) m -> Cstr.sat c m.
  Proof.
    unfold fromCstr; unfoldDebug; intros c m.
    destruct (Shadow.addp Shadow.top c) as [p' [fwd bwd]|ce].
    - apply try_catch_decomp.
      PolCsimplify.
    - apply if_decomp.
      PolCsimplify.
  Qed.

(** Test whether a polyhedron is included in a given half space. *)
  Definition require (c: Cstr.t) (p: t): bool :=
    isIncl p (fromCstr c None).

  Local Opaque Cs.chkIncl isIncl.
  Local Hint Resolve fromCstr_precise isIncl_correct: polC.

  Lemma require_correct: forall c a, If require c a THEN forall m, gamma a m -> Cstr.sat c m.
  Proof.
    intros c a.
    unfold require.
    PolCsimplify.
  Qed.

End CstrPol.


(* a naive (and inefficient) implementation of rename *)
Module NaiveRename (Import D: HasAssume QNum QCond BasicPolQ) <: HasRename QNum BasicPolQ.

  Import BasicPolQ.

  Definition rename (x y: PVar.t) (a:t) : t :=
    project (assume (Eq (Term.Var y) (Term.Var x)) a) x.

  Lemma rename_correct: forall x y a, isRename gamma Program.Basics.impl (rename x y a) a x y.
  Proof.
    unfold isRename, Basics.impl, rename.
    intros; PolCsimplify.
    intros; eapply gamma_ext; [ auto | idtac | eapply project_correct; apply assume_correct; simpl; [idtac | eauto] ].
    + intros x'; autorewrite with progvar.
      erewrite @Mem.assign_id; auto.
    +  autorewrite with progvar.
       unfold Mem.assign. case (PVar.eq_dec x y); auto.
  Qed. 

End NaiveRename.

(* implementation of rename *)
Module Rename (Import D: HasAssume QNum QCond BasicPolQ) <: HasRename QNum BasicPolQ.

  Import BasicPolQ.

  Definition rename (x y: PVar.t) (a:t) : t :=
  DO p <- a -;
  Some {| cons := Cs.rename x y (cons p) ;
          ml := Shadow.rename x y (ml p) |}.

  Lemma rename_correct: forall x y a, isRename gamma Program.Basics.impl (rename x y a) a x y.
  Proof.
    unfold isRename, Basics.impl, rename.
    intros; PolCsimplify.
    rewrite Cs.rename_correct.
    auto.
  Qed. 

End Rename.



(* Gluing all this together *)
Module PolLinQ <: AbstractDomain QNum QCond 
  := BasicPolQ <+ AddIsBottom QNum <+ QCondAssume BasicPolQ CstrPol
  <+ QCondHasAtomicRequire BasicPolQ CstrPol <+ AffCondRequire QNum QCond
  <+ PolQItv
  <+ Rename.

Module FullPolLinQ <: FullItvAbstractDomain QNum QCond QItv 
  := MakeAssign QNum QCond PolLinQ PolLinQ <+ AssignLiftItv QNum QCond QItv PolLinQ PolLinQ PolQItv
                                           <+ AssignSimplePrinting QNum QCond PolLinQ PolLinQ PolLinQ
  .
