(** This module offers an abstract domain of polyhedra
which takes as inputs linear integer expressions ([LinExpr.ZExpr]). *)

Require String.
Require Import Itv.
Require Import PolC.

Import AssignD.
Import OptionMonad.
Import ZCond.

Module CorePolLinZ.

  Definition t:= PolLinQ.t.

  Definition gamma (a:t) m
    := PolLinQ.gamma a (Mem.lift ZtoQ.ofZ m).

(* NO MORE NEEDED FOR BoundVarD !
  Instance mdo: MayDependOn := {| mayDependOn:=PolLinQ._mayDependOn |}.
  Hint Immediate PolLinQ.gamma_mdo: polC.

  Instance gamma_mdo: mdoExt gamma iff.
  Proof.
    assert (mdoExt gamma Program.Basics.impl); auto with progvar.
    constructor 1.
    - unfold gamma, Basics.impl. intros; erewrite eval_mdo_compat; eauto with polC.
    unfold bExt, Mem.lift in * |- *. intros; rewrite H; auto.
    - unfold gamma, Basics.impl. intros; erewrite eval_diff_compat; eauto with polC.
    unfold bExt, Mem.lift in * |- *. intros; rewrite H0; auto.
  Qed.
*)
   Instance gamma_ext: Proper (Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> iff) gamma.
   Proof.
     unfold gamma.
     intros x y H m1 m2 H0. subst; rewrite H0; tauto.
   Qed.

  Definition isIncl:=PolLinQ.isIncl.
  Opaque PolLinQ.isIncl.

  Lemma isIncl_correct: forall a1 a2, If isIncl a1 a2 THEN forall m, gamma a1 m -> gamma a2 m.
  Proof.
    unfold gamma. 
    xsimplify ltac:(eauto with vpl).
  Qed.

  Definition top:= PolLinQ.top.

  Lemma top_correct: forall m, gamma top m.
    simpl; auto.
  Qed.
  Hint Resolve top_correct: vpl.

  Definition bottom: t := PolLinQ.bottom.

  Lemma bottom_correct: forall m, ~(gamma bottom m).
  Proof.
    simpl; auto.
  Qed.

  Definition widen:= PolLinQ.widen.

  Definition join:= PolLinQ.join.

  Lemma join_correct1 p1 p2 m: gamma p1 m -> gamma (join p1 p2) m.
  Proof.
    unfold gamma; auto with vpl.
  Qed.

  Lemma join_correct2 p1 p2 m: gamma p2 m -> gamma (join p1 p2) m.
  Proof.
    unfold gamma; auto with vpl.
  Qed.


  Definition assume (c:ZCond.t) (a:t) : t :=
    PolLinQ.assume (QCond.lift c) a.

  Lemma assume_correct: forall c a m, sat c m -> gamma a m -> gamma (assume c a) m.
  Proof.
    unfold assume, gamma.
    intros; apply PolLinQ.assume_correct; auto.
    autorewrite with vpl.
    auto.
  Qed.

  Definition project (a:t) (x:PVar.t) : t := 
    PolLinQ.project a x.

  Lemma project_correct: forall a x m v, gamma a m -> gamma (project a x) (Mem.assign x v m).
  Proof.
    unfold project, gamma. intros.
    eapply PolLinQ.gamma_ext; [ auto | idtac | eapply PolLinQ.project_correct; eauto].
    autorewrite with vpl; eauto.
  Qed.

  Definition rename (x y: PVar.t) (a:t): t :=
    PolLinQ.rename x y a.

  Lemma rename_correct: forall x y a, isRename gamma Basics.impl (rename x y a) a x y.
  Proof.
    intros x y a.
    unfold gamma,  isRename, Basics.impl.
    intros; eapply PolLinQ.rename_correct.
    eapply PolLinQ.gamma_ext; [ auto | idtac | eauto].
    autorewrite with vpl; auto.
  Qed.

  Definition pr: t -> String.string
    := fun ab => PolLinQ.pr ab.

End CorePolLinZ.


Module PolZItv <: HasGetItv ZNum ZTerm ZItv CorePolLinZ.

  Import CorePolLinZ.

  Definition get_itv: ZTerm.t -> t -> ZItv.t
    := fun a ab =>
      let qa := QTerm.lift a in
        let qitv := PolLinQ.get_itv qa ab in
          proj1_sig (ZItv.fromQItv qitv).

  Lemma get_itv_correct: forall (a: ZTerm.t) (rho: Mem.t ZNum.t) (ab: t),
    gamma ab rho ->
    ZItv.sat (get_itv a ab) (ZTerm.eval a rho).
  Proof.
    intros a rho ab hab.
    unfold get_itv.
    apply (proj2_sig (ZItv.fromQItv _)).
    rewrite <- QTerm.lift_correct.
    apply PolLinQ.get_itv_correct.
    assumption.
  Qed.

End PolZItv.



Module WeakAbsPolLinZ<: WeakAbstractDomain ZNum ZCond
  := CorePolLinZ <+ AddIsBottom ZNum.



Module ZAtomicRequire <: HasAtomicRequire ZNum ZCond WeakAbsPolLinZ.

Import WeakAbsPolLinZ.

Definition atomicRequire (cmp:AffCond.cmpT) (tl tr:ZTerm.t) (a:t) : option bool :=
  PolLinQ.atomicRequire cmp (QTerm.lift tl) (QTerm.lift tr) a.

Local Hint Resolve PolLinQ.atomicRequire_correct: vpl.

Local Opaque ZCond.sat QCond.sat.

Lemma atomicRequire_correct: 
  forall t1 t2 cmp a, WHEN b <- atomicRequire cmp t1 t2 a THEN If b THEN forall m, gamma a m -> ZCond.sat (ZCond.Atom cmp t1 t2) m.
Proof.
  VPLsimplify.
  unfold gamma; intros.
  rewrite <- QCond.lift_correct.
  simpl; auto.
Qed.

End ZAtomicRequire.


Module PolLinZ <: AbstractDomain ZNum ZCond 
  := WeakAbsPolLinZ <+ ZAtomicRequire <+ AffCondRequire ZNum ZCond <+ PolZItv.

Module FullPolLinZ <: FullItvAbstractDomain ZNum ZCond ZItv
  := MakeAssign ZNum ZCond PolLinZ PolLinZ <+ AssignLiftItv ZNum ZCond ZItv PolLinZ PolLinZ PolZItv 
                                           <+ AssignSimplePrinting ZNum ZCond PolLinZ PolLinZ PolLinZ
  .

(* 
(* HERE: see remark in PolC. particular case of "require". *)

Definition GetItvCstrSpec (e: ZExpr.t) (b: bool) (ab: A): Prop
  := forall m: ZMem.t,
       gamma ab m -> ZExpr.eval m e = (if b then ZNum.u else ZNum.z).

Definition getItvCstr (e: ZExpr.t) (b: bool) (ab: A): Exc (GetItvCstrSpec e b ab).
  refine (
      let (qe, pfe) := QExpr.ofZExpr e in
      match PolLinQ.getItvCstr qe b ab with
        | value pf => value _
        | error => error
      end).
  unfold GetItvCstrSpec.
  unfold QExpr.OfZExprSpec in pfe.
  unfold PolLinQ.GetItvCstrSpec in pf.
  intros m h.
  destruct b.
  - apply ZExpr.EvalDenoteT, pfe.
    apply QExpr.EvalDenoteT, pf.
    assumption.
  - apply ZExpr.EvalDenoteF.
    intro hcontr.
    refine ((proj1 (QExpr.EvalDenoteF (Mem.liftZMem m) qe)) (pf _ _) _).
    assumption.
    apply pfe.
    assumption.
Defined.
*)
