open BinNums
open BinPos
open Bool
open Datatypes
open Peano

module PositiveOrderedTypeBits = 
 struct 
  type t = positive
  
  (** val eqb : positive -> positive -> bool **)
  
  let eqb =
    Pos.eqb
  
  (** val eq_dec : positive -> positive -> bool **)
  
  let eq_dec x y =
    let b = Pos.eqb x y in if b then true else false
  
  (** val compare : positive -> positive -> comparison **)
  
  let rec compare x y =
    match x with
    | Coq_xI x0 ->
      (match y with
       | Coq_xI y0 -> compare x0 y0
       | _ -> Gt)
    | Coq_xO x0 ->
      (match y with
       | Coq_xO y0 -> compare x0 y0
       | _ -> Lt)
    | Coq_xH ->
      (match y with
       | Coq_xI y0 -> Lt
       | Coq_xO y0 -> Gt
       | Coq_xH -> Eq)
 end

module PositiveSet = 
 struct 
  module E = PositiveOrderedTypeBits
  
  type elt = positive
  
  type tree =
  | Leaf
  | Node of tree * bool * tree
  
  type t = tree
  
  (** val empty : tree **)
  
  let empty =
    Leaf
  
  (** val is_empty : t -> bool **)
  
  let rec is_empty = function
  | Leaf -> true
  | Node (l, b, r) ->
    if if negb b then is_empty l else false then is_empty r else false
  
  (** val mem : positive -> t -> bool **)
  
  let rec mem i = function
  | Leaf -> false
  | Node (l, o, r) ->
    (match i with
     | Coq_xI i0 -> mem i0 r
     | Coq_xO i0 -> mem i0 l
     | Coq_xH -> o)
  
  (** val add : positive -> t -> t **)
  
  let rec add i = function
  | Leaf ->
    (match i with
     | Coq_xI i0 -> Node (Leaf, false, (add i0 Leaf))
     | Coq_xO i0 -> Node ((add i0 Leaf), false, Leaf)
     | Coq_xH -> Node (Leaf, true, Leaf))
  | Node (l, o, r) ->
    (match i with
     | Coq_xI i0 -> Node (l, o, (add i0 r))
     | Coq_xO i0 -> Node ((add i0 l), o, r)
     | Coq_xH -> Node (l, true, r))
  
  (** val singleton : positive -> t **)
  
  let singleton i =
    add i empty
  
  (** val node : tree -> bool -> tree -> tree **)
  
  let node l b r =
    if b
    then Node (l, b, r)
    else (match l with
          | Leaf ->
            (match r with
             | Leaf -> Leaf
             | Node (t0, b0, t1) -> Node (l, false, r))
          | Node (t0, b0, t1) -> Node (l, false, r))
  
  (** val remove : positive -> t -> t **)
  
  let rec remove i = function
  | Leaf -> Leaf
  | Node (l, o, r) ->
    (match i with
     | Coq_xI i0 -> node l o (remove i0 r)
     | Coq_xO i0 -> node (remove i0 l) o r
     | Coq_xH -> node l false r)
  
  (** val union : t -> t -> t **)
  
  let rec union m m' =
    match m with
    | Leaf -> m'
    | Node (l, o, r) ->
      (match m' with
       | Leaf -> m
       | Node (l', o', r') ->
         Node ((union l l'), (if o then true else o'), (union r r')))
  
  (** val inter : t -> t -> tree **)
  
  let rec inter m m' =
    match m with
    | Leaf -> Leaf
    | Node (l, o, r) ->
      (match m' with
       | Leaf -> Leaf
       | Node (l', o', r') ->
         node (inter l l') (if o then o' else false) (inter r r'))
  
  (** val diff : t -> t -> tree **)
  
  let rec diff m m' =
    match m with
    | Leaf -> Leaf
    | Node (l, o, r) ->
      (match m' with
       | Leaf -> m
       | Node (l', o', r') ->
         node (diff l l') (if o then negb o' else false) (diff r r'))
  
  (** val equal : t -> t -> bool **)
  
  let rec equal m m' =
    match m with
    | Leaf -> is_empty m'
    | Node (l, o, r) ->
      (match m' with
       | Leaf -> is_empty m
       | Node (l', o', r') ->
         if if eqb o o' then equal l l' else false then equal r r' else false)
  
  (** val subset : t -> t -> bool **)
  
  let rec subset m m' =
    match m with
    | Leaf -> true
    | Node (l, o, r) ->
      (match m' with
       | Leaf -> is_empty m
       | Node (l', o', r') ->
         if if if negb o then true else o' then subset l l' else false
         then subset r r'
         else false)
  
  (** val rev_append : positive -> positive -> positive **)
  
  let rec rev_append y x =
    match y with
    | Coq_xI y0 -> rev_append y0 (Coq_xI x)
    | Coq_xO y0 -> rev_append y0 (Coq_xO x)
    | Coq_xH -> x
  
  (** val rev : positive -> positive **)
  
  let rev x =
    rev_append x Coq_xH
  
  (** val xfold : (positive -> 'a1 -> 'a1) -> t -> 'a1 -> positive -> 'a1 **)
  
  let rec xfold f m v i =
    match m with
    | Leaf -> v
    | Node (l, b, r) ->
      if b
      then xfold f r (f (rev i) (xfold f l v (Coq_xO i))) (Coq_xI i)
      else xfold f r (xfold f l v (Coq_xO i)) (Coq_xI i)
  
  (** val fold : (positive -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1 **)
  
  let fold f m i =
    xfold f m i Coq_xH
  
  (** val xforall : (positive -> bool) -> t -> positive -> bool **)
  
  let rec xforall f m i =
    match m with
    | Leaf -> true
    | Node (l, o, r) ->
      if if if negb o then true else f (rev i)
         then xforall f r (Coq_xI i)
         else false
      then xforall f l (Coq_xO i)
      else false
  
  (** val for_all : (positive -> bool) -> t -> bool **)
  
  let for_all f m =
    xforall f m Coq_xH
  
  (** val xexists : (positive -> bool) -> t -> positive -> bool **)
  
  let rec xexists f m i =
    match m with
    | Leaf -> false
    | Node (l, o, r) ->
      if if if o then f (rev i) else false
         then true
         else xexists f r (Coq_xI i)
      then true
      else xexists f l (Coq_xO i)
  
  (** val exists_ : (positive -> bool) -> t -> bool **)
  
  let exists_ f m =
    xexists f m Coq_xH
  
  (** val xfilter : (positive -> bool) -> t -> positive -> tree **)
  
  let rec xfilter f m i =
    match m with
    | Leaf -> Leaf
    | Node (l, o, r) ->
      node (xfilter f l (Coq_xO i)) (if o then f (rev i) else false)
        (xfilter f r (Coq_xI i))
  
  (** val filter : (positive -> bool) -> t -> tree **)
  
  let filter f m =
    xfilter f m Coq_xH
  
  (** val xpartition : (positive -> bool) -> t -> positive -> tree*tree **)
  
  let rec xpartition f m i =
    match m with
    | Leaf -> Leaf,Leaf
    | Node (l, o, r) ->
      let lt,lf = xpartition f l (Coq_xO i) in
      let rt,rf = xpartition f r (Coq_xI i) in
      if o
      then let fi = f (rev i) in (node lt fi rt),(node lf (negb fi) rf)
      else (node lt false rt),(node lf false rf)
  
  (** val partition : (positive -> bool) -> t -> tree*tree **)
  
  let partition f m =
    xpartition f m Coq_xH
  
  (** val xelements : t -> positive -> positive list -> positive list **)
  
  let rec xelements m i a =
    match m with
    | Leaf -> a
    | Node (l, b, r) ->
      if b
      then xelements l (Coq_xO i) ((rev i)::(xelements r (Coq_xI i) a))
      else xelements l (Coq_xO i) (xelements r (Coq_xI i) a)
  
  (** val elements : t -> positive list **)
  
  let elements m =
    xelements m Coq_xH []
  
  (** val cardinal : t -> nat **)
  
  let rec cardinal = function
  | Leaf -> O
  | Node (l, b, r) ->
    if b
    then S (plus (cardinal l) (cardinal r))
    else plus (cardinal l) (cardinal r)
  
  (** val omap : (elt -> elt) -> elt option -> elt option **)
  
  let omap f = function
  | Some i -> Some (f i)
  | None -> None
  
  (** val choose : t -> positive option **)
  
  let rec choose = function
  | Leaf -> None
  | Node (l, o, r) ->
    if o
    then Some Coq_xH
    else (match choose l with
          | Some i -> Some (Coq_xO i)
          | None -> omap (fun x -> Coq_xI x) (choose r))
  
  (** val min_elt : t -> positive option **)
  
  let rec min_elt = function
  | Leaf -> None
  | Node (l, o, r) ->
    (match min_elt l with
     | Some i -> Some (Coq_xO i)
     | None ->
       if o then Some Coq_xH else omap (fun x -> Coq_xI x) (min_elt r))
  
  (** val max_elt : t -> positive option **)
  
  let rec max_elt = function
  | Leaf -> None
  | Node (l, o, r) ->
    (match max_elt r with
     | Some i -> Some (Coq_xI i)
     | None ->
       if o then Some Coq_xH else omap (fun x -> Coq_xO x) (max_elt l))
  
  (** val compare_bool : bool -> bool -> comparison **)
  
  let compare_bool a b =
    if a then if b then Eq else Gt else if b then Lt else Eq
  
  (** val compare : t -> t -> comparison **)
  
  let rec compare m m' =
    match m with
    | Leaf -> if is_empty m' then Eq else Lt
    | Node (l, o, r) ->
      (match m' with
       | Leaf -> if is_empty m then Eq else Gt
       | Node (l', o', r') ->
         (match compare_bool o o' with
          | Eq ->
            (match compare l l' with
             | Eq -> compare r r'
             | x -> x)
          | x -> x))
  
  (** val eq_dec : t -> t -> bool **)
  
  let eq_dec s s' =
    if equal s s' then true else false
 end

