type cmpT
= Eq | Le | Lt

type t = {
	typ: cmpT;
	v: Vec.t;
	c: Num.t }

let get_typ (x : t) = x.typ
let get_v (x : t) = x.v
let get_c (x : t) = x.c

type cstr_t = t

type prop_t =
| Trivial
| Contrad
| Nothing

exception BadMult
exception CannotElim
exception NoElim

let cmpAdd: cmpT -> cmpT -> cmpT
= fun o1 o2 ->
	match o1 with
	| Eq -> o2
	| Lt as o -> o
	| Le as o ->
		match o2 with
		| Eq | Le -> o
		| Lt as o -> o

let eval c pt: bool =
	let r = Num.cmp (Rtree.fold Num.add Num.z (Vec.mult_t c.v pt)) c.c in
		match c.typ with Eq -> r = 0 | Le -> r <= 0 | Lt -> r < 0

let add c1 c2 = {
	typ = cmpAdd c1.typ c2.typ;
	v = Vec.add c1.v c2.v;
	c = Num.add c1.c c2.c }

let mult n c = { c with v = Vec.mult n c.v; c = Num.mult n c.c }

let elim c1 c2 va =
	let a1 = Vec.get c1.v va in
        let a2 = Vec.get c2.v va in
	let r1 = Num.cmpz a1 in
        let r2 = Num.cmpz a2 in
	match (r1 = 0, r2 = 0) with
	| (true, true) -> raise NoElim
	| (true, false) | (false, true) -> raise CannotElim
	| (false, false) ->
		let (n1, n2) = 
			match (c1.typ, c2.typ) with
			| (Eq, Eq) -> (Num.neg a2, a1)
			| (Le, Eq) | (Lt, Eq) ->
				if r2 > 0 then (Num.neg a2, a1) else (a2, Num.neg a1)

			| (Eq, Le) | (Eq, Lt) ->
				if r1 > 0 then (a2, Num.neg a1) else (Num.neg a2, a1)

			| Le, Le | Lt, Lt
			| Lt, Le | Le, Lt ->
				match (r1 > 0, r2 > 0) with
				| (true, true) | (false, false) -> raise CannotElim
				| (true, false) -> (a2, Num.neg a1)
				| (false, true) -> (Num.neg a2, a1)
		in
		let c = add (mult n1 c1) (mult n2 c2) in
		let gcd = Vec.gcd c.v in
		(mult gcd c, Num.mult gcd n1, Num.mult gcd n2)

let compl (c: cstr_t): cstr_t =
	match c.typ with
	| Eq -> invalid_arg "Cstr.compl"
	| Le -> { typ = Lt; v = Vec.neg c.v; c = Num.neg c.c }
	| Lt -> { typ = Le; v = Vec.neg c.v; c = Num.neg c.c }

let split c =
	match c.typ with
	| Le | Lt -> invalid_arg "Cstr.split"
	| Eq -> ({ typ = Le; v = c.v; c = c.c },
		{ typ = Le; v = Vec.neg c.v; c = Num.neg c.c })

let mk typ v c = { typ = typ; v = Vec.mk v; c = c }
let eq v c = mk Eq v c
let le v c = mk Le v c
let lt v c = mk Lt v c

let mkInt typ v c =
	match typ with
	| Eq | Le -> { typ = typ; v = Vec.mk v; c = c }
	| Lt -> { typ = Le; v = Vec.mk v; c = Num.sub c Num.u }

let isEqSyn c1 c2 =
	c1.typ = c2.typ && Vec.isEq c1.v c2.v && Num.cmp c1.c c2.c = 0

let isInclSyn c1 c2 =
	if Vec.isEq c1.v c2.v then
		match c1.typ, c2.typ with
		| Eq, Eq -> Num.cmp c1.c c2.c = 0
		| Eq, Le | Le, Le
		| Lt, Lt | Lt, Le -> Num.cmp c1.c c2.c <= 0
		| Eq, Lt | Le, Lt -> Num.cmp c1.c c2.c < 0
		| Le, Eq | Lt, Eq -> false
	else
		false

let isEq c1 c2 =
	if c1.typ <> c2.typ then
		false
	else

	match Vec.isomorph c1.v c2.v with
	| None ->
		if Vec.isEq Vec.nil c1.v && Vec.isEq Vec.nil c2.v then
			Num.cmp c1.c c2.c = 0
		else
			false

	| Some ratio ->
		if Num.cmpz ratio > 0 then
			if c1.typ = Eq then
				Num.cmp c1.c (Num.mult ratio c2.c) = 0
			else
				false
		else
			Num.cmp c1.c (Num.mult ratio c2.c) = 0

let isIncl c1 c2 =
	match c1.v, c2.v with
	| _, Rtree.Nil ->
		let r = Num.cmpz c2.c in
		(match c2.typ with
		| Eq -> r = 0
		| Le -> r <= 0
		| Lt -> r < 0)

	| Rtree.Nil, _ -> false
	| Rtree.Sub _, Rtree.Sub _ ->
		match Vec.isomorph c1.v c2.v with
		| None -> false
		| Some ratio ->
			let res = Num.cmp c1.c (Num.mult ratio c2.c) in
			match c1.typ, c2.typ with
			| Eq, Eq -> res = 0
			| Le, Le | Lt, Lt
			| Lt, Le -> Num.cmpz ratio < 0 && res <= 0
			| Le, Lt -> Num.cmpz ratio < 0 && res < 0
			| Eq, Le -> res <= 0
			| Eq, Lt -> res < 0
			| Le, Eq | Lt, Eq -> false

let tellProp c =
	if Vec.isEq Vec.nil c.v then
		let r = Num.cmpz c.c in
		match c.typ with
		| Eq -> if r = 0 then Trivial else Contrad
		| Le -> if r <= 0 then Trivial else Contrad
		| Lt -> if r < 0 then Trivial else Contrad
	else
		Nothing

let listVars: t list -> Var.t list
= fun l -> Vec.listVars (List.map get_v l)

let getCoefsFor : Var.t option -> t list -> Num.t list
= function
	| None -> fun l -> List.map get_c l
	| Some x -> fun l -> List.map (fun c -> Vec.get (get_v c) x) l

let pr: (Var.t -> string) -> t -> string
= fun varPr c ->
	let sign =
		match c.typ with
		| Eq -> " = "
		| Le -> " <= "
		| Lt -> " < "
	in
	(Vec.pr varPr c.v) ^ sign ^ (Num.pr c.c)
