let (>:) : 'a -> ('a -> 'b) -> 'b
= fun x f -> f x

let (-:) : ('a -> 'b) -> ('b -> 'c) -> 'a -> 'c
= fun f g x -> g (f x)

module M
= struct

	type 'a t
	= 'a option

	let unit : 'a -> 'a t
	= fun x -> Some x

	let bind : 'a t -> ('a -> 'b t) -> 'b t
	= function None -> fun _ -> None | Some a -> fun f -> f a

	let comp : ('a -> 'b t) -> ('b -> 'c t) -> 'a -> 'c t
	= fun f g x -> bind (f x) g

	let outKill : string -> 'a t -> 'a
	= fun e -> function None -> Pervasives.failwith e | Some x -> x

end

let (@~) = M.bind
let (@=) = M.comp

let arrayFold2 : ('a -> 'b -> 'c -> 'a) -> 'a -> 'b array -> 'c array -> 'a option
= fun f init a a' ->
	let l = Array.length a in
	if Array.length a' <> l
	then None
	else
		let r : 'a ref = ref init in
		for i = 0 to l - 1
		do r := f !r a.(i) a'.(i);
		done; Some !r

let arrayFoldI : ('a -> int -> 'b -> 'a) -> 'a -> 'b array -> 'a
= fun f init a ->
	let r : 'a ref = ref init in
	for i = 0 to Array.length a - 1
	do r := f !r i a.(i);
	done; !r

let binOp : ('a -> 'b -> 'a) -> 'a array -> 'b array -> 'a array option
= fun f a a' ->
	let l = Array.length a in
	if Array.length a' <> l
	then None
	else
		(for i = 0 to l - 1
		do a.(i) <- f a.(i) a'.(i);
		done; Some a)

let unOp : ('a -> 'a) -> 'a array -> 'a array
= fun f a ->
	for i = 0 to Array.length a - 1
	do a.(i) <- f a.(i);
	done; a

let toBool : bool option -> bool
= function None -> false | Some x -> x

let killOnFail : string -> 'a option -> 'a
= fun e -> function Some x -> x | None -> Pervasives.failwith e

type paramT
= int

let paramPr : paramT -> string
= fun d -> "d" ^ (Pervasives.string_of_int d)

type pCoefT
= Num.t array

let pCoefPr : pCoefT -> string
=	let pr1 : int -> Num.t -> string
	= function
		| 0 -> Num.pr
		| i -> fun a -> Num.pr a ^ "*" ^ paramPr i
	in
	fun c ->
		Array.to_list c >:
		List.mapi pr1 -:
		String.concat " + "

let pCoefEq : pCoefT -> pCoefT -> bool
= fun pc pc' ->
	arrayFold2 (fun r a a' -> r && Num.cmp a a' = 0) true pc pc' >:
	toBool

let pCoefAdd : pCoefT -> pCoefT -> pCoefT
= fun pc pc' ->
	binOp Num.add pc pc' >:
	killOnFail "Psplx.pCoefAdd"

let pCoefMul : pCoefT -> Num.t -> pCoefT
= fun pc n -> Array.iteri (fun i a -> pc.(i) <- Num.mult a n) pc; pc

let pCoefIsZ : pCoefT -> bool
= fun pc -> Array.fold_left (fun r a -> r && Num.isZ a) true pc

type pValT
= Num.t array

let pValPr : pValT -> string
=	let pr1 : int -> Num.t -> string
	= function
		| 0 -> fun a -> "_ -> " ^ Num.pr a
		| i -> fun a -> paramPr i ^ " -> " ^ Num.pr a
	in
	fun v ->
		Array.to_list v >:
		List.mapi pr1 -:
		String.concat "; "

let pValEq : pValT -> pValT -> bool
= fun pv pv' ->
	arrayFold2 (fun r a a' -> r && Num.cmp a a' = 0) true pv pv' >:
	toBool

type varT
= int

let varPr : varT -> string
= fun x -> "l" ^ (Pervasives.string_of_int x)

let varEq : varT -> varT -> bool
= (=)

type pObjT
= pCoefT array

let pObjPr : pObjT -> string
= fun o ->
	let s : string ref
	= ref (Printf.sprintf "%s*(%s)" (varPr 0) (pCoefPr o.(0)))
	in
	for i = 1 to Array.length o - 1
	do s := Printf.sprintf "%s + %s*(%s)" !s (varPr i) (pCoefPr o.(i));
	done;
	!s

let pObjEq : pObjT -> pObjT -> bool
= fun po po' ->
	arrayFold2 (fun r a a' -> r && pCoefEq a a') true po po' >:
	toBool

let pObjAdd : pObjT -> pObjT -> pObjT
= fun po po' ->
	binOp pCoefAdd po po' >:
	killOnFail "Psplx.pObjAdd"

let pObjComb : pObjT list -> pObjT
= function
	| [] -> Pervasives.invalid_arg "Psplx.pObjComb"
	| po::l ->
		let r = Array.copy po in
		List.fold_left (fun r po -> pObjAdd r po) r l

let pObjCoef : varT -> pObjT -> pCoefT
= fun x po -> po.(x)

type vecT
= Num.t array

let vecPr : vecT -> string
= fun v ->
	Array.sub v 0 (Array.length v - 1) >:
	Array.to_list -:
	List.mapi (fun i a -> Num.pr a ^ "*" ^ varPr i) -:
	String.concat " + " -:
	fun s -> Printf.sprintf "%s = %s" s (Num.pr v.(Array.length v - 1))

let vecEq : vecT -> vecT -> bool
= fun v v' ->
	arrayFold2 (fun r a a' -> r && Num.cmp a a' = 0) true v v' >:
	toBool

let vecAdd : vecT -> vecT -> vecT
= fun v v' ->
	binOp Num.add v v' >:
	killOnFail "Psplx.vecAdd"

let vecMul : vecT -> Num.t -> vecT
= fun v n -> unOp (fun a -> Num.mult a n) v

let vecComb : (Num.t * vecT) list -> vecT
= function
	| [] -> Pervasives.invalid_arg "Psplx.vecComb"
	| (n, v)::l ->
		let r = Array.copy (vecMul v n) in
		List.fold_left (fun r (n, v) -> vecAdd r (vecMul v n)) r l

let vecCoefMul : pCoefT -> vecT -> pObjT
= fun pc v ->
	Array.init (Array.length v) (fun i -> pCoefMul (Array.copy pc) v.(i))

let pCoefEval : pValT -> pCoefT -> Num.t
= fun pv pc ->
	arrayFold2 (fun r a b -> Num.add r (Num.mult a b)) Num.z pc pv >:
	killOnFail "Psplx.pCoefEval"

let pObjEval : pValT -> pObjT -> vecT
= fun pv po -> Array.map (fun pc -> pCoefEval pv pc) po

type consT
= {xb : varT; v: vecT}

let consPr : consT -> string
= fun c -> Printf.sprintf "%s : %s" (varPr c.xb) (vecPr c.v)

let consEq : consT -> consT -> bool
= fun c c' -> varEq c.xb c'.xb && vecEq c.v c'.v

let consCoef : varT -> consT -> Num.t
= fun x c -> c.v.(x)

let consCst : consT -> Num.t
= fun c -> c.v.(Array.length c.v - 1)

type tabT
= {obj : pObjT; pval : pValT; cs : consT array}

let tabPr : tabT -> string
= fun t ->
	let hdr = Printf.sprintf "obj: %s\nwith %s" (pObjPr t.obj) (pValPr t.pval) in
	Array.fold_left (fun s c -> s ^ "\n" ^ consPr c) hdr t.cs

let tabEq : tabT -> tabT -> bool
=	let consAEq : consT array -> consT array -> bool
	= fun a a' ->
		arrayFold2 (fun r c c' -> r && consEq c c') true a a' >:
		toBool
	in
	fun t t' -> pObjEq t.obj t'.obj
		&& pValEq t.pval t'.pval && consAEq t.cs t'.cs

let tabNbCons : tabT -> int
= fun t -> Array.length t.cs

let tabDim : tabT -> int
= fun t -> Array.length t.obj - 1

let tabCoef : int -> varT -> tabT -> Num.t
= fun i x t -> consCoef x t.cs.(i)

let tabDepVars : tabT -> varT list
= fun tab ->
	Array.fold_left (fun l c -> c.xb :: l) [] tab.cs >:
	List.rev

module VarSet
= Set.Make (Var)

let setOfList : Var.t list -> VarSet.t
= fun xl -> List.fold_left (fun s x -> VarSet.add x s) VarSet.empty xl

let constrainedSet : (int * Cstr.t) list -> VarSet.t
= fun l -> setOfList (Cstr.listVars (List.map Pervasives.snd l))

let buildCons : (int * Cstr.t) list -> Var.t option -> vecT
= fun l mx ->
	let al = Cstr.getCoefsFor mx (List.map Pervasives.snd l) in
	let c = Array.make (List.length al + 1) Num.z in
	List.iteri (fun i a -> c.(i) <- a) al; c

let projCons : (int * Cstr.t) list -> Var.t -> vecT
= fun l x -> buildCons l (Some x)

let normCons : (int * Cstr.t) list -> Cstr.t -> vecT
= fun l c ->
	let xl = Vec.toList (Cstr.get_v c) in
	let xl' = (Cstr.get_c c, None) :: List.map (fun (n, x) -> (n, Some x)) xl in
	let comb = List.map (fun (n, mx) -> n, buildCons l mx) xl' in
	let v' = vecComb comb in
	v'.(Array.length v' - 1) <- Num.u; v'

let buildObj : (int * Cstr.t) list -> Var.t list -> pObjT * pValT
=	let buildTerm : (int * Cstr.t) list -> int -> int -> Var.t -> pObjT
	= fun l len i x ->
		let a = Array.make (len + 1) Num.z in
		a.(i+1) <- Num.u;
		vecCoefMul a (buildCons l (Some x))
	in
	let initPVal : VarSet.t -> pValT
	=	VarSet.cardinal -: 
		fun n ->  Array.init (n + 1) (function 0 -> Num.z | _ -> Num.u)
	in
	let build : (int * Cstr.t) list -> VarSet.t -> pObjT
	= fun l xs ->
		VarSet.elements xs >:
		List.mapi (buildTerm l (VarSet.cardinal xs)) -:
		pObjComb -:
		unOp (unOp Num.neg)
	in 
	fun l xl ->
		let xs = VarSet.diff (constrainedSet l) (setOfList xl) in
		build l xs, initPVal xs

let tabMk : (int * Cstr.t) list -> Var.t list -> Cstr.t -> tabT
= fun l xl c ->
	let cl = normCons l c :: List.map (projCons l) xl in
	let obj, pval = buildObj l xl in {
		obj = obj;
		pval = pval;
		cs = Array.init (List.length cl) (fun i -> {xb = -1; v = List.nth cl i})
	}

let vecNorm : varT -> vecT -> vecT
= fun x v ->
	let a = Num.inv v.(x) in
	vecMul v a

let consProj : varT -> consT -> consT -> consT
= fun x nc c ->
	let a = Num.neg c.v.(x) in
	{c with v = vecAdd c.v (vecMul (Array.copy nc.v) a)}

let pivotCs : int -> varT -> consT array -> consT array
= fun line x a ->
	let nc = {xb = x; v = vecNorm x a.(line).v} in
	Array.iteri (fun i c -> a.(i) <- if i = line then nc else consProj x nc c) a; a

let pivotObj : varT -> consT -> pObjT -> pObjT
= fun x c o ->
	Array.copy c.v >:
	unOp Num.neg -:
	vecCoefMul o.(x) -:
	pObjAdd o

let pivotMain : int -> varT -> tabT -> tabT option
= fun line col tab ->
	let cs' = pivotCs line col tab.cs in
	Some {tab with obj = pivotObj col cs'.(line) tab.obj; cs = cs'}

let lineInRange : int -> tabT -> tabT option
= fun i t -> if 0 <= i && i < tabNbCons t then Some t else None

let colInRange : varT -> tabT -> tabT option
= fun i t -> if 0 <= i && i < tabDim t then Some t else None

let pivotM : int -> varT -> tabT -> tabT M.t
= fun line col tab ->
	M.unit tab @~
	lineInRange line @=
	colInRange col @=
	pivotMain line col

let pivot : int -> varT -> tabT -> tabT
= fun line col tab ->
	pivotM line col tab >:
	M.outKill "Psplx.pivot"

let chkDep : varT -> tabT -> bool
=	let chkCs : varT -> consT array -> bool option
	= fun x a -> Array.fold_left (function
		| None -> fun _ -> None
		| Some r -> fun c ->
			if Num.isZ (consCoef x c)
			then Some r else if r then None else Some true) (Some false) a
	in
	fun x t -> pCoefIsZ (pObjCoef x t.obj) && chkCs x t.cs = Some true

let chooseLast : tabT -> (int * varT) list option
= fun t ->
	if not (chkDep (tabDim t - 1) t)
	then None
	else
		if Num.isZ (tabCoef 0 (tabDim t - 1) t)
		then None
		else Some [0, tabDim t - 1]

let choose1 : (int * varT) list -> int -> consT -> (int * varT) list option
= fun l i c ->
	let r : varT option ref = ref None in
	let f : varT -> Num.t -> unit
	= fun x a ->
		match !r with
		| Some _ -> ()
		| None ->
			if Num.isZ a
			then ()
			else
				try Pervasives.ignore (List.find (fun (_, x') -> varEq x x') l)
				with Not_found -> r := Some x
	in
	try Pervasives.ignore (List.find (fun (i', _) -> i = i') l); Some l
	with Not_found ->
		Array.iteri f c.v;
		match !r with
		| None -> None
		| Some x -> Some ((i, x) :: l)

let chooseInitDep : tabT -> (int * varT) list option
= fun t ->
	let ml : (int * varT) list option ref = ref (chooseLast t) in
	let f : int -> consT -> unit
	= fun i c ->
		match !ml with
		| None -> ()
		| Some l -> ml := choose1 l i c
	in
	Array.iteri f t.cs; !ml

let tabInitM : tabT -> tabT M.t
= fun t ->
	chooseInitDep t @~
	List.fold_left (fun t' (i, x) -> t' @~ pivotM i x) (M.unit t)

let tabInit : tabT -> tabT
= fun t ->
	tabInitM t >:
	M.outKill "Psplx.tabInit"

type strategyT
= {
	pickIndep : tabT -> varT option;
	pickDep : varT -> tabT -> int option
}

module Bland
= struct

	let pickIndep : tabT -> varT option
	=	let choose : varT option -> int -> Num.t -> varT option
		= function
			| Some _ as x -> fun _ _ -> x
			| None -> fun x a ->
				if Num.cmpz a > 0 then Some x else None
		in
		fun tab ->
			pObjEval tab.pval tab.obj >:
			arrayFoldI choose None 

	let pickDep : varT -> tabT -> int option
	=	let score : varT -> int -> consT -> Num.t option * int
		= fun x i c ->
			let ax = consCoef x c in
			if Num.cmpz ax < 0
			then Some (Num.div (consCst c) ax), i
			else None, i
		in
		let min : Num.t option * int -> Num.t option * int -> Num.t option * int
		= fun (mb, i) (mb', i') ->
			match mb with
			| None -> mb', i'
			| Some b ->
				match mb' with
				| None -> mb, i
				| Some b' -> if Num.cmp b b' <= 0 then mb, i else mb', i'
		in
		let f : varT -> Num.t option * int -> int -> consT -> Num.t option * int
		= fun x r i c -> min r (score x i c)
		in
		fun x tab ->
			let mb, i = arrayFoldI (f x) (None, -1) tab.cs in
			match mb with
			| None -> None
			| Some _ -> Some i

	let strgy : strategyT
	= {pickIndep = pickIndep; pickDep = pickDep}

end

let rec min : strategyT -> tabT -> tabT
= fun strgy tab ->
	match strgy.pickIndep tab with
	| None -> tab
	| Some x ->
		match strgy.pickDep x tab with
		| None -> tab
		| Some i -> min strgy (pivot i x tab)

let nextDelta : pObjT -> pValT -> int -> Num.t
=	let numMax : Num.t -> Num.t -> Num.t
	= fun n n' -> if Num.cmp n n' < 0 then n' else n
	in
	let f : int -> Num.t -> Num.t -> pCoefT -> Num.t -> Num.t
	= fun i deltaI nval pc objv ->
		let deltaCoef = pc.(i) in
		if Num.cmpz deltaCoef >= 0
		then nval
		else
			let ch = Num.div objv deltaCoef in
			numMax nval (Num.sub deltaI ch)
	in
	fun po pv i ->
		arrayFold2 (f i pv.(i)) Num.negU po (pObjEval pv po) >:
		M.outKill "Psplx.nextDelta"
