(** the type of parameters in the parametric objective *)
type paramT
= int

val paramPr : paramT -> string

(** a parametric coefficient in the parametric objective, indexed by [paramT].
The first element (at index 0) of a [pCoefT] stores the constant
part and the coefficient of parameter \delta_i (i>0) is stored in
the ith element. *)
type pCoefT
= Num.t array

val pCoefPr : pCoefT -> string
val pCoefEq : pCoefT -> pCoefT -> bool

(** [pCoefAdd pc pc'] adds [pc] to [pc'] and stores the result in [pc].
The return value is a reference to [pc]. *)
val pCoefAdd : pCoefT -> pCoefT -> pCoefT

(** [pCoefMul pc n] stores the result of the multiplication of [pc] by [n] in [pc].
The result is a reference to [pc]. *)
val pCoefMul : pCoefT -> Num.t -> pCoefT

(** an assignment of a value to parameters, indexed by [paramT].
The first element of a [pCoefT] contains the constant part of the coeffient,
its value should always be [Num.z] in a [pValT]. *)
type pValT
= Num.t array

val pValPr : pValT -> string
val pValEq : pValT -> pValT -> bool

(** the type of the variables of the problem *)
type varT
= int

val varPr : varT -> string
val varEq : varT -> varT -> bool

(** the type of the parametric objective, indexed by [varT] *)
type pObjT
= pCoefT array

val pObjPr : pObjT -> string
val pObjEq : pObjT -> pObjT -> bool

(** [pObjAdd po po'] adds two values of type [pObjT] and stores the result in [po].
The result in a reference to [po]. *)
val pObjAdd : pObjT -> pObjT -> pObjT

(** Sum a list of [pObjT] and store the result in a new array. *)
val pObjComb : pObjT list -> pObjT

(** the type of constraint of the simplex tableau, indexed by [varT].
The last element containts the constant. *)
type vecT
= Num.t array

val vecPr : vecT -> string
val vecEq : vecT -> vecT -> bool

(** [vecAdd v v'] stores the result of the addition of [v] and [v'] in [v].
The result is a reference to [v]. *)
val vecAdd : vecT -> vecT -> vecT

(** [vecMul v n] stores the result of the multiplication of [v] by [n] in [v].
The result is a reference to [v]. *)
val vecMul : vecT -> Num.t -> vecT

(** [vecComb l] builds the linear combination described by [l].
The result is stored in a new array. Upon return of the function,
easy of the [vecT] in [l] has been multiplied by its coefficient. *)
val vecComb : (Num.t * vecT) list -> vecT

(** [vecCoefMul pc v] builds the result of the multiplication of [v]
by [pc] in a newly allocated [pObjT]. *)
val vecCoefMul : pCoefT -> vecT -> pObjT

(** constraints with annotations about their dependent variable *)
type consT
= {xb : varT; v: vecT}

val consPr : consT -> string
val consEq : consT -> consT -> bool

(** the type of a simplex tableau *)
type tabT
= {obj : pObjT; pval : pValT; cs : consT array}

val tabPr : tabT -> string
val tabEq : tabT -> tabT -> bool

(** Get the list of dependent variables in the given tableau,
in the order in which they appear. *)
val tabDepVars : tabT -> varT list

(** [projCons l x] builds the constraint of the tableau which guarantees
that variable [x] is projected. *)
val projCons : (int * Cstr.t) list -> Var.t -> vecT

(** [normCons l c] builds the normalization constraint specified by point [c]. *)
val normCons : (int * Cstr.t) list -> Cstr.t -> vecT

(** [buildObj l xl] builds the parametric objective function for the projection
of variables [xl] as well as the initial assignment of values to parameters. *)
val buildObj : (int * Cstr.t) list -> Var.t list -> pObjT * pValT

(** [mk l xl c] builds the initial simplex tableau for the projection of the
variables [xl]. The point/hyperplane used for constraint normalization is given
as [c].
No initial choice of basic variables is made. *)
val tabMk : (int * Cstr.t) list -> Var.t list -> Cstr.t -> tabT

(** [tabInit t] initialises tableau [t], meaning that initial dependent
variables are chosen and necessary pivoting is performed. *)
val tabInit : tabT -> tabT

(** [pivot line col tab] makes variable [col] dependent in line [line]
of tableau [tab]. Memory is shared between [tab] and the result: [tab]
should not be used after a call to [pivot]. *)
val pivot : int -> varT -> tabT -> tabT

(** The type of an iteration strategy in the simplex optimization. *)
type strategyT
= {
	pickIndep : tabT -> varT option;
	pickDep : varT -> tabT -> int option
}

(** Iteration strategy which follows Bland's rule. *)
module Bland : sig val strgy : strategyT end

(** Perform a minimization on the given tableau,
according to the given strategy. The tableau should
be initialized with [tabInit]. *)
val min : strategyT -> tabT -> tabT

val nextDelta : pObjT -> pValT -> int -> Num.t
