let samples: Db.t = List.fold_left Db.add Db.nil Data.db

module Key =
struct
	type t = { op: Db.opT; sz: int }

	let compare k1 k2 =
		let p = Db.opCmp k1.op k2.op in
		if p <> 0 then p else Pervasives.compare k1.sz k2.sz
end

module Stat = Map.Make ( Key )

type valueT = { apron: int64; libpoly: int64; check: int64; pt: int }
let valueZ: valueT = { apron = Int64.zero; libpoly = Int64.zero; check = Int64.zero; pt = 0 }

type statT = valueT Stat.t
let statZ = Stat.empty

let update (o: Db.opT) (v: valueT) (s: Db.sampleT): valueT =
	let avg (l: int64 list): int64 =
		match l with [] -> failwith "F.update"
		| _::_ ->
			Int64.div (List.fold_left Int64.add Int64.zero l) (Int64.of_int (List.length l))
	in
	match o with
	| Db.Incl | Db.Join -> {
		apron = Int64.add v.apron (avg s.Db.apron);
		libpoly = Int64.add v.libpoly (avg s.Db.libpoly);
		check = Int64.add v.check (avg s.Db.check);
		pt = v.pt + 1 }
	| Db.Meet | Db.Assign -> {
		apron = Int64.add v.apron (avg s.Db.apron);
		libpoly = Int64.add v.libpoly (avg s.Db.libpoly);
		check = v.check;
		pt = v.pt + 1 }

let add (m: Db.metaT) (s: Db.sampleT) (stat: statT): statT =
	let sz = List.fold_left (+) 0 (s.Db.eqs @ s.Db.ineqs) in
	let k = { Key.op = m.Db.op; Key.sz = sz } in
	let v = try Stat.find k stat with Not_found -> valueZ in
	Stat.add k (update k.Key.op v s) stat

let stat: statT = Db.M.fold add samples statZ

let prUnit n = 
	let u = ["ns"; "us"; "ms"; "s"] in
	let prD d64 =
		let d = Int64.to_int d64 in
		if d >= 100 then
			"." ^ (string_of_int d)
		else if d >= 10 then
			".0" ^ (string_of_int d)
		else if d > 0 then
			".00" ^ (string_of_int d)
		else
			""
	in
	let rec pr i d = function
	| [] -> (Int64.to_string i) ^ (prD d) ^ "min"
	| h::t ->
		let ni = Int64.div i (Int64.of_int 1000) in
		if Int64.compare Int64.zero ni = 0 then
			(Int64.to_string i) ^ (prD d) ^ h
		else
			let nd = Int64.rem i (Int64.of_int 1000) in
			pr ni nd t
	in
	pr n Int64.zero u

let valueAdd v1 v2 = {
	apron = Int64.add v1.apron v2.apron;
	libpoly = Int64.add v1.libpoly v2.libpoly;
	check = Int64.add v1.check v2.check;
	pt = v1.pt + v2.pt }

let select (n: int): int =
	if n = 0 || n = 1 then 0
	else if n > 30 then 7
	else (n + 4) / 5

let labels = [| "0-1"; "2-5"; "6-10"; "11-15"; "16-20"; "21-25"; "26-30"; "31+" |]
let outputIncl = ref (Array.make 8 valueZ)
let outputJoin = ref (Array.make 8 valueZ)
let outputMeet = ref (Array.make 8 valueZ)
let outputAssign = ref (Array.make 8 valueZ)

let prOut s a =
	Array.iteri (fun i v -> Printf.printf "%s, %s, apron:%s, libpoly:%s, data points: %d\n"
		s (Array.get labels i) (prUnit v.apron) (prUnit v.libpoly) v.pt) !a
;;

Stat.iter (fun (k: Key.t) (v: valueT) ->
	let a =
		match k.Key.op with
		| Db.Incl -> outputIncl
		| Db.Join -> outputJoin
		| Db.Meet -> outputMeet
		| Db.Assign -> outputAssign
	in
	let i = select k.Key.sz in
	let o = Array.get !a i in
	Array.set !a i (valueAdd o v)) stat;

prOut "incl" outputIncl;
prOut "join" outputJoin;
prOut "meet" outputMeet;
prOut "assign" outputAssign


