open QArith_base

let x = Var.fromInt 1
let y = Var.fromInt 2
let z = Var.fromInt 3
let a = Var.fromInt 4

let varPr: Var.t -> string
= fun _x ->
	let vars: (Var.t * string) list
	= [x, "x"; y, "y"; z, "z"; a, "a"]
	in
	List.assoc _x vars

let mk l = List.map (fun (i, a) -> (i, Num.mk1 a)) l

let mkc t v c =
  Cstr.mk t (List.map (fun (c, v) -> (Num.mk1 c, v)) v) (Num.mk1 c)

let eq = mkc Cstr.Eq
let le = mkc Cstr.Le
let lt = mkc Cstr.Lt

let joinEltPr: Cert.joinEltT -> string
= fun j -> Printf.sprintf "{id = %i;\narg1:\n%s\narg2:\n%s\n}"
	(Cert.getId j) (Cert.prC (Cert.getArg1 j)) (Cert.prC (Cert.getArg2 j))

let joinPr: Cert.joinT -> string
= fun j -> String.concat "\n" (List.map joinEltPr j)

(* Glue.isInclC *)
let isInclCTs: T.testT
=
	let chk (name, res, p1, p2) = fun state ->
		if Glue.isInclC p1 p2 = res
		then T.succeed state
		else
			match Pol.isIncl p1 p2 with
			| Pol.NoIncl -> T.fail name "bad result" state
			| Pol.Incl c ->
				let e = Printf.sprintf
		"trying to prove inclusion of p1 in p2 with c\np1:\n%s\np2:\n%s\nc:\n%s"
					(Pol.prExt varPr p1) (Pol.prExt varPr p2) (Cert.pr c)
				in
				T.fail name e state
	in
	T.suite "incl" (List.map chk Pol_t.isInclTcs)

(* Glue.addC *)
let addCTs: T.testT
=
	let chk (name, isEmpty, c, p, _) = fun state ->
	    try 
	      match Glue.addC p c with
		| None ->
			if isEmpty then
				T.succeed state
			else
				T.fail name "unexpectedly empty" state
		| Some p1 ->
			if isEmpty then
				T.fail name "should be empty" state
			else
				if Glue.polChkSync (Some p1) then
					T.succeed state
				else
	   				T.fail name (CoqPr.charListTr (PolC.PolLinQ.pr (Some p1))) state
             with
	     | Failure _ ->
			let ce =
				match Pol.add p c with
				| Pol.Added (_, (ce, _)) -> ce
				| Pol.Contrad ce -> ce
			in
			let estr =
				Printf.sprintf "error in the checker.\ncertificate:\n%s\n"
					(Cert.pr ce)
			in
			T.fail name estr state

	in
	T.suite "addC" (List.map chk Pol_t.addTcs)

(* ConsSet.Cs.addp *)
(* XXX: this probably should not be here *)
let addpTs: T.testT
=	let chk (nm, _, c, p, _)
	= fun st ->
		let cc = Support.cToCstrC c in
		match Support.addp p cc with
		| Shadow.Contrad _ -> T.succeed st (* XXX: nothing checked *)
		| Shadow.Added (_, (fwd, bwd)) ->
			let pc = Support.pToConsSet p in
			try
				match ConsSet.Cs.addp pc (Pol.get_nxt p) cc fwd bwd with
				| Some _ -> T.succeed st
				| None ->
					let e = "checker failed" in
					T.fail nm e st
			with Failure e -> T.fail nm e st
	in
	T.suite "addp" (List.map chk Pol_t.addTcs)

(* Glue.meet *)
let meetTs: T.testT
=	let chk (nm, p1, p2, maybeP)
	=	let noneExpected: PolC.PolLinQ.t -> T.stateT -> T.stateT
		= function
			| None -> fun st -> T.succeed st
			| Some _ as p -> fun st ->
				let e = Printf.sprintf "not empty\np:\n%s\n"
					(CoqPr.charListTr (PolC.PolLinQ.pr p))
				in
				T.fail nm e st
		in
		let someExpected: PolC.PolLinQ.t -> T.stateT -> T.stateT
		= function
			| None -> fun st -> T.fail nm "got None" st
			| Some _ as p -> fun st ->
				if Glue.polChkSync p
				then T.succeed st
				else
					let e = Printf.sprintf "out of sync\np:\n%s\n"
						(CoqPr.charListTr (PolC.PolLinQ.pr p))
					in
					T.fail nm e st
		in
		fun st ->
			try
				let res = Glue.meet p1 p2 in
				match maybeP with
				| None -> noneExpected res st
				| Some _ -> someExpected res st
			with Failure msg ->
				let e = "checker failure " ^ msg in
				T.fail nm e st
	in
	T.suite "meet" (List.map chk Pol_t.meetTcs)

(* Glue.joinC *)
let joinCTs: T.testT
=
	let chk (name, p1, p2, _) = fun state ->
	  try
		match Glue.joinC p1 p2 with
		| None             -> T.fail name "Bottom" state
		| Some p           ->
			if Glue.polChkSync (Some p) then
				T.succeed state
			else
				T.fail name (CoqPr.charListTr (PolC.PolLinQ.pr (Some p))) state
	  with
	    | Failure _ -> 
			let (p, ce) = Pol.join p1 p2 in
			let e = Printf.sprintf
				"Glue.joinC returned None\ncert:\n%s\np1:\n%s\np2:\n%s\nres:\n%s\n"
				(joinPr ce) (Pol.prExt varPr p1) (Pol.prExt varPr p2)
				(Pol.prExt varPr p)
			in
			T.fail name e state

	in
	T.suite "join" (List.map chk Pol_t.joinTcs)

(* Glue.getItvC *)
let getItvCTs
=
	let chk (name, v, itv, p) state =
		let itv1 = Glue.getItvC p v in
		if Pol_t.eqItv itv itv1 then
			T.succeed state
		else
			T.fail name (Pol.prItv varPr v itv1) state
	in
	T.suite "getItvC" (List.map chk Pol_t.itvizeTcs)

(* Glue.project *)
let projectTs: T.testT
=	(* Checking synchronization between the two components of the result is enough.
	We know the Ocaml result is correct, since we are usings the same test cases
	as in Pol_t.projectTs. *)
	let chk (nm, x, p, r): T.stateT -> T.stateT
	= fun st ->
	    try
	      let p'=Glue.project p x in
			if Glue.polChkSync p'
			then T.succeed st
			else
				let e = "out of sync" in
				T.fail nm e st
	    with
		| Failure _ -> T.fail nm "got None" st
	in
	T.suite "project" (List.map chk Pol_t.projectTcs)

(* Glue.projectM *)
let projectMTs: T.testT
=
	let chk (nm, l, p, _)
	= fun st ->
		try
			let _ = Glue.projectM p l in
			T.succeed st
		with Failure "checking error" -> T.fail nm "got Failure" st
	in
	T.suite "projectM" (List.map chk Pol_t.projectMTcs)

let ts: T.testT
= T.suite "Glue" [
	isInclCTs;
	addCTs;
	addpTs;
	meetTs;
	joinCTs;
	getItvCTs;
	projectTs;
	projectMTs
]
