(* Test suite for Splx.Opt. *)
let x = Var.fromInt 1
let y = Var.fromInt 2
let z = Var.fromInt 3
let t = Var.fromInt 4

let nxt = Var.fromInt 5

let varPr: Var.t -> string
= fun x ->
	let vars = [x, "x"; y, "y"; z, "z"; t, "t"] in
	try
		List.assoc x vars
	with
	| Not_found -> "v" ^ (Var.pr x)

let mkc t v c =
	Cstr.mk t (List.map (fun (c, v) -> (Num.mk1 c, v)) v) (Num.mk1 c)

let sxLift : Splx.t Splx.mayUnsatT -> Splx.t
= function
	| Splx.IsOk sx -> sx
	| Splx.IsUnsat _ -> Pervasives.invalid_arg "Opt_t.sxLift"

let eq = mkc Cstr.Eq
let le = mkc Cstr.Le
let lt = mkc Cstr.Lt

let eqProgress b1 b2 =
	match b1, b2 with
	| Splx.Opt.Unbnd, Splx.Opt.Unbnd -> true
	| Splx.Opt.NoChange, Splx.Opt.NoChange -> true
	| Splx.Opt.UpTo n1, Splx.Opt.UpTo n2 -> Splx.Val.cmp n1 n2 = 0
	| _, _ -> false

let eqNext n1 n2 =
	match n1, n2 with
	| Splx.Opt.GoOn _, Splx.Opt.GoOn _ -> failwith "Opt_t.eqNext: unimplemented case"
	| Splx.Opt.OptUnbnd, Splx.Opt.OptUnbnd -> true
	| Splx.Opt.OptFinite v1, Splx.Opt.OptFinite v2 -> Splx.Val.cmp v1 v2 = 0
	| Splx.Opt.GoOn _, Splx.Opt.OptUnbnd
	| Splx.Opt.GoOn _, Splx.Opt.OptFinite _
	| Splx.Opt.OptUnbnd, Splx.Opt.GoOn _
	| Splx.Opt.OptUnbnd, Splx.Opt.OptFinite _
	| Splx.Opt.OptFinite _, Splx.Opt.OptUnbnd
	| Splx.Opt.OptFinite _, Splx.Opt.GoOn _ -> false

let eqOpt o1 o2 =
	match o1, o2 with
	| Splx.Opt.Finite (n1, w1), Splx.Opt.Finite (n2, w2)
	| Splx.Opt.Sup (n1, w1), Splx.Opt.Sup (n2, w2) -> Num.cmp n1 n2 = 0 && Cert.isEqF w1 w2
	| Splx.Opt.Infty, Splx.Opt.Infty -> true
	| Splx.Opt.Finite (_, _), Splx.Opt.Sup (_, _)
	| Splx.Opt.Finite (_, _), Splx.Opt.Infty
	| Splx.Opt.Sup (_, _), Splx.Opt.Finite (_, _)
	| Splx.Opt.Sup (_, _), Splx.Opt.Infty
	| Splx.Opt.Infty, Splx.Opt.Finite (_, _)
	| Splx.Opt.Infty, Splx.Opt.Sup (_, _) -> false

(* The invariant on the simplex tableau which is maintained during the optimization is stronger
than that of the satisfiability algorithm.
All the constraints of the constraint matrix should be satisfied and
all the variables, basic or non-basic, should be within their bounds. *)
let chkInv s0 =
	let chkBnds s =
		let chkBnd st =
			(match Splx.get_low st with
			| Some l -> Splx.Val.cmp (Splx.get_bv l) (Splx.get_v st) <= 0
			|None -> true) &&
			(match Splx.get_up st with
			| Some u -> Splx.Val.cmp (Splx.get_bv u) (Splx.get_v st) >= 0
			|None -> true)
		in
		Rtree.fold (fun a n -> a && (chkBnd n)) true (Splx.get_state s) in
	let chkMat s =
		let chkCons maybeC =
			let rec eval a ve st =
				match ve, st with
				| Rtree.Nil, _ | _, Rtree.Nil-> a
				| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
					let incr = Splx.Val.add a (Splx.Val.muln (Splx.get_v n2) n1) in
					eval (eval incr l1 l2) r1 r2
			in
			match maybeC with
			| None -> true
			| Some c -> Splx.Val.cmp Splx.Val.z (eval Splx.Val.z c (Splx.get_state s)) = 0
		in
		Rtree.fold (fun a n -> a && (chkCons n)) true (Splx.get_mat s)
	in
	chkMat s0 && chkBnds s0

(* Splx.Opt.setObj *)
let setObjTs: T.testT
=
	let chkVar (t, obj, s) = fun state ->
		let (oVar, s1) = Splx.Opt.setObj (sxLift s) obj in
		match Rtree.get None (Splx.get_mat s1) oVar with
		| None -> T.fail t "no constraint associated with z" state
		| Some cons -> (* This is a light check. *)
			if Num.cmp (Vec.get cons oVar) Num.u = 0 then
				T.succeed state
			else
				T.fail t "not equal" state
	in
	let chkInv (t, obj, s) = fun state ->
		let (_, s1) = Splx.Opt.setObj (sxLift s) obj in
		if chkInv s1 then
			T.succeed state
		else
			let msg = Printf.sprintf "invariant\n%s" (Splx.pr varPr s1) in
			T.fail t msg state
	in
	let chkBnd (t, obj, s) = fun state ->
		let (oVar, s1) = Splx.Opt.setObj (sxLift s) obj in
		let oSt = Rtree.get {Splx.v = Splx.Val.z; Splx.low = None; Splx.up = None} (Splx.get_state s1) oVar in
		if (Splx.get_low oSt) = None && (Splx.get_up oSt) = None then
			T.succeed state
		else
			T.fail t "bad bounds" state
	in
	let tcs = [
		"nil0", Vec.nil, Splx.mk nxt [];
		"nil1", Vec.mk [Num.u, x], Splx.mk nxt [];
		"nil2", Vec.mk [Num.u, x; Num.u, y], Splx.mk nxt [];
		"u0", Vec.mk [Num.u, x], Splx.mk nxt [ 0, le [1, x] 1 ];
		"u1", Vec.mk [Num.u, x], Splx.mk nxt [ 0, le [-1, x] (-1) ];
		"m0", Vec.mk [Num.u, x; Num.u, y], Splx.mk nxt [
			0, eq [1, y] 3;
			1, le [-1, x] (1) ]
	] in
	T.suite "setObj" [
		T.suite "var" (List.map chkVar tcs);
		T.suite "invariant" (List.map chkInv tcs);
		T.suite "bnd" (List.map chkBnd tcs)
	]

(* Splx.Opt.pickNBasic *)
let pickNBasicTs: T.testT
=
	let chk (t, a, (o, s)) = fun state ->
		let isMoveEq (v1, d1, b1) (v2, d2, b2) = eqProgress b1 b2 && v1 = v2 && d1 = d2 in
		match Splx.Opt.pickNBasic o s with
		| Splx.Opt.Done ->
			if a = None then
				T.succeed state
			else
				T.fail t "not equal to None" state
		| Splx.Opt.Move (v, d, b) as r ->
			let msg = Printf.sprintf "pickNBasic found %s" (Splx.Opt.prAction varPr r) in
			match a with
			| None -> T.fail t msg state
			| Some aL ->
				if List.exists (isMoveEq (v, d, b)) aL then
					T.succeed state
				else
					T.fail t msg state
	in
	let tcs = [
		"nilV0", None, Splx.Opt.setObj (sxLift (Splx.mk nxt []))
			Vec.nil;

		"nilV1", None, Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, le [1, x; 1, y] 2 ]))
			Vec.nil;

		"nilS0", Some [
				x, Splx.Opt.Incr, Splx.Opt.Unbnd
			], Splx.Opt.setObj (sxLift (Splx.mk nxt []))
			(Vec.mk [Num.u, x]);

		"nilS1", Some [
				x, Splx.Opt.Incr, Splx.Opt.Unbnd;
				y, Splx.Opt.Incr, Splx.Opt.Unbnd
			], Splx.Opt.setObj (sxLift (Splx.mk nxt []))
			(Vec.mk [Num.u, x; Num.u, y]);

		"nilS2", Some [
				x, Splx.Opt.Decr, Splx.Opt.Unbnd;
				y, Splx.Opt.Incr, Splx.Opt.Unbnd
			], Splx.Opt.setObj (sxLift (Splx.mk nxt []))
			(Vec.mk [Num.negU, x; Num.u, y]);

		"bnd0", Some [
				x, Splx.Opt.Incr, Splx.Opt.UpTo (Splx.Val.of_num (Num.mk1 2))
			], Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, le [1, x] 2 ]))
			(Vec.mk [Num.u, x]);

		"bnd1", Some [
				x, Splx.Opt.Incr, Splx.Opt.Unbnd
			], Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, le [1, x; 1, y] 2;
				1, le [1, y] 1 ]))
			(Vec.mk [Num.u, x]);

		"ubnd0", Some [
				x, Splx.Opt.Decr, Splx.Opt.Unbnd
			], Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, le [1, x] 2 ]))
			(Vec.mk [Num.negU, x]);

		"notFirst0", Some [
				y, Splx.Opt.Incr, Splx.Opt.Unbnd
			], Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, eq [1, x] 0 ]))
			(Vec.mk [Num.u, x; Num.u, y]);

		"notFirst1", Some [
				y, Splx.Opt.Incr, Splx.Opt.UpTo (Splx.Val.of_num (Num.mk1 2))
			], Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, eq [1, x] 0;
				1, le [1, y] 2 ]))
			(Vec.mk [Num.u, x; Num.u, y]);

		"none0", None, Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, eq [1, x] 0;
				1, le [1, y] (-1) ]))
			(Vec.mk [Num.u, x; Num.u, y])
	] in
	T.suite "pickNBasic" (List.map chk tcs)

(* Splx.Opt.pickBasic *)
let pickBasicTs: T.testT
=
	let chkV (t, xN, inBnd, dir, xB, _, (_, s)) = fun state ->
		let (xB1, _) = Splx.Opt.pickBasic s xN inBnd dir in
		if xB = xB1 then
			T.succeed state
		else
			T.fail t (varPr xB1) state
	in
	let chkP (t, xN, inBnd, dir, _, xNBnd, (_, s)) = fun state ->
		let (_, xNBnd1) = Splx.Opt.pickBasic s xN inBnd dir in
		if eqProgress xNBnd xNBnd1 then
			T.succeed state
		else
			T.fail t (Splx.Opt.prProgress xNBnd1) state
	in
	let sx1 = nxt in
	let sx2 = Var.next sx1 in
	let tcs = [
		"nil0", x, Splx.Opt.Unbnd, Splx.Opt.Incr, x, Splx.Opt.Unbnd,
			Splx.Opt.setObj (sxLift (Splx.mk nxt []))
			(Vec.mk [Num.u, x]);

		"up0", x, Splx.Opt.Unbnd, Splx.Opt.Decr, sx1, Splx.Opt.UpTo (Splx.Val.of_num (Num.mk1 3)),
			Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, le [-1, x; 1, y] 3 ]))
			(Vec.mk [Num.u, x]);

		"up1", x, Splx.Opt.Unbnd, Splx.Opt.Decr, sx1, Splx.Opt.UpTo (Splx.Val.of_num (Num.mk1 2)),
			Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, le [-2, x; 1, y] 4 ]))
			(Vec.mk [Num.u, x]);

		"up2", x, Splx.Opt.Unbnd, Splx.Opt.Decr, sx2, Splx.Opt.NoChange,
			Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, le [-2, x; 1, y] 4;
				1, le [-1, x; 1, z] 0 ]))
			(Vec.mk [Num.u, x]);

		"nbasic0", x, Splx.Opt.UpTo (Splx.Val.of_num (Num.mk1 1)), Splx.Opt.Decr,
			x, Splx.Opt.UpTo (Splx.Val.of_num (Num.mk1 1)),
			Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, le [-2, x; 1, y] 4 ]))
			(Vec.mk [Num.u, x])
	] in
	T.suite "pickBasic" [
		T.suite "var" (List.map chkV tcs);
		T.suite "progress" (List.map chkP tcs)
	]

(* Splx.Opt.step *)
let stepEndTs: T.testT
=
	let chk (t, next, (z, s)) = fun state ->
		let next1 = Splx.Opt.step z s in
		match next with
		| Splx.Opt.GoOn _ -> T.fail t "bad test case" state
		| _ ->
			if eqNext next next1 then
				T.succeed state
			else
				T.fail t (Splx.Opt.prNext varPr next1) state
	in
	let tcs = [
		"nil0", Splx.Opt.OptUnbnd, Splx.Opt.setObj (sxLift (Splx.mk nxt []))
			(Vec.mk [Num.u, x]);

		"u0", Splx.Opt.OptFinite (Splx.Val.of_num (Num.mk1 (-2))),
			Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, le [-1, x] (-2) ]))
			(Vec.mk [Num.negU, x])
	] in
	T.suite "end" (List.map chk tcs)

let stepGoOnTs: T.testT
=
	let chkInv (t, _, (z, s)) = fun state ->
		match Splx.Opt.step z s with
		| Splx.Opt.OptFinite _
		| Splx.Opt.OptUnbnd as r -> T.fail t (Splx.Opt.prNext varPr r) state
		| Splx.Opt.GoOn s1 ->
			if chkInv s1 then
				T.succeed state
			else
				let msg = Printf.sprintf "broken invariant\n%s\n" (Splx.pr varPr s1) in
				T.fail t msg state
	in
	let chkObj (t, zV, (z, s)) = fun state ->
		match Splx.Opt.step z s with
		| Splx.Opt.OptFinite _
		| Splx.Opt.OptUnbnd as r -> T.fail t (Splx.Opt.prNext varPr r) state
		| Splx.Opt.GoOn s1 ->
			let zV1 =
				let zSt = Rtree.get {Splx.v = Splx.Val.z; Splx.low = None; Splx.up = None} (Splx.get_state s1) z in
				Splx.get_v zSt
			in
			if Splx.Val.cmp zV zV1 = 0 then
				T.succeed state
			else
				T.fail t (Splx.Val.pr zV1) state
	in
	let tcs = [
		"own0", Splx.Val.of_num (Num.mk1 2), Splx.Opt.setObj (sxLift (Splx.mk nxt [
				0, le [1, x] 2 ]))
			(Vec.mk [Num.u, x]);

		"other0", Splx.Val.of_num (Num.mk1 2), Splx.Opt.setObj
			(sxLift (Splx.mk nxt [ 0, le [1, x; 1, y] 2 ]))
			(Vec.mk [Num.u, x])
	] in
	T.suite "goOn" [
		T.suite "inv" (List.map chkInv tcs);
		T.suite "obj" (List.map chkObj tcs)
	]

let stepTs: T.testT
= T.suite "step" [stepEndTs; stepGoOnTs]

(* Splx.Opt.max *)
let maxTs: T.testT
=
	let chk (t, v, obj, s) = fun state ->
		match Splx.Opt.max' s obj with
		| Splx.IsUnsat _ -> T.fail t "unsat" state
		| Splx.IsOk opt ->
			if eqOpt v opt
			then T.succeed state
			else T.fail t (Splx.Opt.prOpt opt) state
	in
	let tcs = [
		"nil0", Splx.Opt.Finite (Num.z, []), Vec.mk [], Splx.mk nxt [];
		"nil1", Splx.Opt.Infty, Vec.mk [Num.u, x], Splx.mk nxt [];
		"own0", Splx.Opt.Finite (Num.mk1 2, [0, Num.u]), Vec.mk [Num.u, x], Splx.mk nxt [
			0, le [1, x] 2 ];

		"ownSup0", Splx.Opt.Sup (Num.mk1 2, [0, Num.u]),
			Vec.mk [Num.u, x], Splx.mk nxt [
				0, lt [1, x] 2 ];

		"comb1", Splx.Opt.Finite (Num.mk1 2, [0, Num.u; 1, Num.u]),
			Vec.mk [Num.u, x], Splx.mk nxt [
				0, le [1, x; 1, y] 2;
				1, le [-1, y] 0 ];

		"comb2", Splx.Opt.Sup (Num.mk1 2, [0, Num.u; 1, Num.u]),
			Vec.mk [Num.u, x], Splx.mk nxt [
				0, le [1, x; 1, y] 2;
				1, lt [-1, y] 0 ];

		"m0", Splx.Opt.Finite (Num.mk1 6, [0, Num.mk 2 1; 1, Num.mk 2 1]),
			Vec.mk [Num.u, x], Splx.mk nxt [
				0, le [-1, y] 4;
				1, le [2, x; 1, y] 8;
				2, le [1, x; 1, y] 3 ];

		"red0", Splx.Opt.Finite (Num.mk1 4, [0, Num.u; 1, Num.u]),
			Vec.mk [Num.u, x; Num.u, y], Splx.mk nxt [
				0, le [1, x] 2;
				1, le [1, y] 2;
				2, le [1, x; 1, y] 5 ];

		"distant_nxt0", Splx.Opt.Infty, Vec.mk [Num.negU, x],
			Splx.mk (Var.XI (Var.XO (Var.XO Var.XH))) [
				1, Cstr.eq [Num.u, x; Num.u, y; Num.negU, z] Num.z;
				2, Cstr.eq [Num.u, y; Num.negU, t] Num.negU;
				3, Cstr.le [Num.negU, z] Num.u;
				4, Cstr.le [Num.negU, z; Num.negU, t] Num.negU]
	] in
	T.suite "max" (List.map chk tcs)

let ts: T.testT
= T.suite "Splx.Opt" [setObjTs; pickNBasicTs; pickBasicTs; stepTs; maxTs]
